"""
This file contains unmodified code from the FilterPy library, specifically the KalmanFilter class.
Original sources:
- https://github.com/rlabbe/filterpy/blob/master/filterpy/kalman/kalman_filter.py
- https://github.com/rlabbe/filterpy/blob/master/filterpy/common/discretization.py
- https://github.com/rlabbe/filterpy/blob/master/filterpy/stats/stats.py

Certain modifications were made to the original code:
- Updated some code to be compatible with newer Python versions and package versions
- Made some simplifications to adjust the code for our purposes
- Refactored some code to be more in line with our codebase style

Original software under MIT license:

Copyright (c) 2015 Roger R. Labbe Jr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

------------------------------------------------------------------------------------

This module implements the linear Kalman filter in both an object
oriented and procedural form. The KalmanFilter class implements
the filter by storing the various matrices in instance variables,
minimizing the amount of bookkeeping you have to do.

All Kalman filters operate with a predict->update cycle. The
predict step, implemented with the method or function predict(),
uses the state transition matrix F to predict the state in the next
time period (epoch). The state is stored as a gaussian (x, P), where
x is the state (column) vector, and P is its covariance. Covariance
matrix Q specifies the process covariance. In Bayesian terms, this
prediction is called the *prior*, which you can think of colloquially
as the estimate prior to incorporating the measurement.

The update step, implemented with the method or function `update()`,
incorporates the measurement z with covariance R, into the state
estimate (x, P). The class stores the system uncertainty in S,
the innovation (residual between prediction and measurement in
measurement space) in y, and the Kalman gain in k. The procedural
form returns these variables to you. In Bayesian terms this computes
the *posterior* - the estimate after the information from the
measurement is incorporated.

Whether you use the OO form or procedural form is up to you. If
matrices such as H, R, and F are changing each epoch, you'll probably
opt to use the procedural form. If they are unchanging, the OO
form is perhaps easier to use since you won't need to keep track
of these matrices. This is especially useful if you are implementing
banks of filters or comparing various KF designs for performance;
a trivial coding bug could lead to using the wrong sets of matrices.

This module also offers an implementation of the RTS smoother, and
other helper functions, such as log likelihood computations.

The Saver class allows you to easily save the state of the
KalmanFilter class after every update

This module expects NumPy arrays for all values that expect
arrays, although in a few cases, particularly method parameters,
it will accept types that convert to NumPy arrays, such as lists
of lists. These exceptions are documented in the method or function.

Examples
--------
The following example constructs a constant velocity kinematic
filter, filters noisy data, and plots the results. It also demonstrates
using the Saver class to save the state of the filter at each epoch.

.. code-block:: Python

    import matplotlib.pyplot as plt
    import numpy as np
    from filterpy.kalman import KalmanFilter
    from filterpy.common import Q_discrete_white_noise, Saver

    r_std, q_std = 2., 0.003
    cv = KalmanFilter(dim_x=2, dim_z=1)
    cv.x = np.array([[0., 1.]]) # position, velocity
    cv.F = np.array([[1, dt],[ [0, 1]])
    cv.R = np.array([[r_std^^2]])
    f.H = np.array([[1., 0.]])
    f.P = np.diag([.1^^2, .03^^2)
    f.Q = Q_discrete_white_noise(2, dt, q_std**2)

    saver = Saver(cv)
    for z in range(100):
        cv.predict()
        cv.update([z + randn() * r_std])
        saver.save() # save the filter's state

    saver.to_array()
    plt.plot(saver.x[:, 0])

    # plot all of the priors
    plt.plot(saver.x_prior[:, 0])

    # plot mahalanobis distance
    plt.figure()
    plt.plot(saver.mahalanobis)

This code implements the same filter using the procedural form

    x = np.array([[0., 1.]]) # position, velocity
    F = np.array([[1, dt],[ [0, 1]])
    R = np.array([[r_std^^2]])
    H = np.array([[1., 0.]])
    P = np.diag([.1^^2, .03^^2)
    Q = Q_discrete_white_noise(2, dt, q_std**2)

    for z in range(100):
        x, P = predict(x, P, F=F, Q=Q)
        x, P = update(x, P, z=[z + randn() * r_std], R=R, H=H)
        xs.append(x[0, 0])
    plt.plot(xs)


For more examples see the test subdirectory, or refer to the
book cited below. In it I both teach Kalman filtering from basic
principles, and teach the use of this library in great detail.

FilterPy library.
http://github.com/rlabbe/filterpy

Documentation at:
https://filterpy.readthedocs.org

Supporting book at:
https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python

This is licensed under an MIT license. See the readme.MD file
for more information.

Copyright 2014-2018 Roger R Labbe Jr.
"""

from copy import deepcopy
import sys
import numpy as np
from scipy.stats import multivariate_normal


class KalmanFilter(object):
    r"""Implements a Kalman filter. You are responsible for setting the
    various state variables to reasonable values; the defaults  will
    not give you a functional filter.

    For now the best documentation is my free book Kalman and Bayesian
    Filters in Python [2]_. The test files in this directory also give you a
    basic idea of use, albeit without much description.

    In brief, you will first construct this object, specifying the size of
    the state vector with dim_x and the size of the measurement vector that
    you will be using with dim_z. These are mostly used to perform size checks
    when you assign values to the various matrices. For example, if you
    specified dim_z=2 and then try to assign a 3x3 matrix to R (the
    measurement noise matrix you will get an assert exception because R
    should be 2x2. (If for whatever reason you need to alter the size of
    things midstream just use the underscore version of the matrices to
    assign directly: your_filter._R = a_3x3_matrix.)

    After construction the filter will have default matrices created for you,
    but you must specify the values for each. It’s usually easiest to just
    overwrite them rather than assign to each element yourself. This will be
    clearer in the example below. All are of type numpy.array.


    Examples
    --------

    Here is a filter that tracks position and velocity using a sensor that only
    reads position.

    First construct the object with the required dimensionality.

    .. code::

        from filterpy.kalman import KalmanFilter
        f = KalmanFilter (dim_x=2, dim_z=1)


    Assign the initial value for the state (position and velocity). You can do this
    with a two dimensional array like so:

        .. code::

            f.x = np.array([[2.],    # position
                            [0.]])   # velocity

    or just use a one dimensional array, which I prefer doing.

    .. code::

        f.x = np.array([2., 0.])


    Define the state transition matrix:

        .. code::

            f.F = np.array([[1.,1.],
                            [0.,1.]])

    Define the measurement function:

        .. code::

        f.H = np.array([[1.,0.]])

    Define the covariance matrix. Here I take advantage of the fact that
    P already contains np.eye(dim_x), and just multiply by the uncertainty:

    .. code::

        f.P *= 1000.

    I could have written:

    .. code::

        f.P = np.array([[1000.,    0.],
                        [   0., 1000.] ])

    You decide which is more readable and understandable.

    Now assign the measurement noise. Here the dimension is 1x1, so I can
    use a scalar

    .. code::

        f.R = 5

    I could have done this instead:

    .. code::

        f.R = np.array([[5.]])

    Note that this must be a 2 dimensional array, as must all the matrices.

    Finally, I will assign the process noise. Here I will take advantage of
    another FilterPy library function:

    .. code::

        from filterpy.common import Q_discrete_white_noise
        f.Q = Q_discrete_white_noise(dim=2, dt=0.1, var=0.13)


    Now just perform the standard predict/update loop:

    while some_condition_is_true:

    .. code::

        z = get_sensor_reading()
        f.predict()
        f.update(z)

        do_something_with_estimate (f.x)


    **Procedural Form**

    This module also contains stand alone functions to perform Kalman filtering.
    Use these if you are not a fan of objects.

    **Example**

    .. code::

        while True:
            z, R = read_sensor()
            x, P = predict(x, P, F, Q)
            x, P = update(x, P, z, R, H)

    See my book Kalman and Bayesian Filters in Python [2]_.


    You will have to set the following attributes after constructing this
    object for the filter to perform properly. Please note that there are
    various checks in place to ensure that you have made everything the
    'correct' size. However, it is possible to provide incorrectly sized
    arrays such that the linear algebra can not perform an operation.
    It can also fail silently - you can end up with matrices of a size that
    allows the linear algebra to work, but are the wrong shape for the problem
    you are trying to solve.

    Parameters
    ----------
    dim_x : int
        Number of state variables for the Kalman filter. For example, if
        you are tracking the position and velocity of an object in two
        dimensions, dim_x would be 4.
        This is used to set the default size of P, Q, and u

    dim_z : int
        Number of of measurement inputs. For example, if the sensor
        provides you with position in (x,y), dim_z would be 2.

    dim_u : int (optional)
        size of the control input, if it is being used.
        Default value of 0 indicates it is not used.

    compute_log_likelihood : bool (default = True)
        Computes log likelihood by default, but this can be a slow
        computation, so if you never use it you can turn this computation
        off.

    Attributes
    ----------
    x : numpy.array(dim_x, 1)
        Current state estimate. Any call to update() or predict() updates
        this variable.

    P : numpy.array(dim_x, dim_x)
        Current state covariance matrix. Any call to update() or predict()
        updates this variable.

    x_prior : numpy.array(dim_x, 1)
        Prior (predicted) state estimate. The *_prior and *_post attributes
        are for convienence; they store the  prior and posterior of the
        current epoch. Read Only.

    P_prior : numpy.array(dim_x, dim_x)
        Prior (predicted) state covariance matrix. Read Only.

    x_post : numpy.array(dim_x, 1)
        Posterior (updated) state estimate. Read Only.

    P_post : numpy.array(dim_x, dim_x)
        Posterior (updated) state covariance matrix. Read Only.

    z : numpy.array
        Last measurement used in update(). Read only.

    R : numpy.array(dim_z, dim_z)
        Measurement noise matrix

    Q : numpy.array(dim_x, dim_x)
        Process noise matrix

    F : numpy.array()
        State Transition matrix

    H : numpy.array(dim_z, dim_x)
        Measurement function

    y : numpy.array
        Residual of the update step. Read only.

    K : numpy.array(dim_x, dim_z)
        Kalman gain of the update step. Read only.

    S :  numpy.array
        System uncertainty (P projected to measurement space). Read only.

    SI :  numpy.array
        Inverse system uncertainty. Read only.

    log_likelihood : float
        log-likelihood of the last measurement. Read only.

    likelihood : float
        likelihood of last measurement. Read only.

        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.

    mahalanobis : float
        mahalanobis distance of the innovation. Read only.

    inv : function, default numpy.linalg.inv
        If you prefer another inverse function, such as the Moore-Penrose
        pseudo inverse, set it to that instead: kf.inv = np.linalg.pinv

        This is only used to invert self.S. If you know it is diagonal, you
        might choose to set it to filterpy.common.inv_diagonal, which is
        several times faster than numpy.linalg.inv for diagonal matrices.

    alpha : float
        Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading
        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon [1]_.

    References
    ----------

    .. [1] Dan Simon. "Optimal State Estimation." John Wiley & Sons.
       p. 208-212. (2006)

    .. [2] Roger Labbe. "Kalman and Bayesian Filters in Python"
       https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python

    """

    def __init__(self, dim_x, dim_z, dim_u=0):
        if dim_x < 1:
            raise ValueError("dim_x must be 1 or greater")
        if dim_z < 1:
            raise ValueError("dim_z must be 1 or greater")
        if dim_u < 0:
            raise ValueError("dim_u must be 0 or greater")

        self.dim_x = dim_x
        self.dim_z = dim_z
        self.dim_u = dim_u

        self.x = np.zeros((dim_x, 1))  # state
        self.P = np.eye(dim_x)  # uncertainty covariance
        self.Q = np.eye(dim_x)  # process uncertainty
        self.B = None  # control transition matrix
        self.F = np.eye(dim_x)  # state transition matrix
        self.H = np.zeros((dim_z, dim_x))  # Measurement function
        self.R = np.eye(dim_z)  # state uncertainty
        self._alpha_sq = 1.0  # fading memory control
        self.M = np.zeros((dim_z, dim_z))  # process-measurement cross correlation
        self.z = np.array([[None] * self.dim_z]).T

        # gain and residual are computed during the innovation step. We
        # save them so that in case you want to inspect them for various
        # purposes
        self.K = np.zeros((dim_x, dim_z))  # kalman gain
        self.y = np.zeros((dim_z, 1))
        self.S = np.zeros((dim_z, dim_z))  # system uncertainty
        self.SI = np.zeros((dim_z, dim_z))  # inverse system uncertainty

        # identity matrix. Do not alter this.
        self._I = np.eye(dim_x)

        # these will always be a copy of x,P after predict() is called
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

        # these will always be a copy of x,P after update() is called
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

        # Only computed only if requested via property
        self._log_likelihood = np.log(sys.float_info.min)
        self._likelihood = sys.float_info.min
        self._mahalanobis = None

        self.inv = np.linalg.inv

    def predict(self, u=None, B=None, F=None, Q=None):
        """
        Predict next state (prior) using the Kalman filter state propagation
        equations.

        Parameters
        ----------

        u : np.array
            Optional control vector. If not `None`, it is multiplied by B
            to create the control input into the system.

        B : np.array(dim_x, dim_z), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.

        F : np.array(dim_x, dim_x), or None
            Optional state transition matrix; a value of None
            will cause the filter to use `self.F`.

        Q : np.array(dim_x, dim_x), scalar, or None
            Optional process noise matrix; a value of None will cause the
            filter to use `self.Q`.
        """

        if B is None:
            B = self.B
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        elif np.isscalar(Q):
            Q = np.eye(self.dim_x) * Q

        # x = Fx + Bu
        if B is not None and u is not None:
            self.x = np.dot(F, self.x) + np.dot(B, u)
        else:
            self.x = np.dot(F, self.x)

        # P = FPF' + Q
        self.P = self._alpha_sq * np.dot(np.dot(F, self.P), F.T) + Q

        # save prior
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def update(self, z, R=None, H=None):
        """
        Add a new measurement (z) to the Kalman filter.

        If z is None, nothing is computed. However, x_post and P_post are
        updated with the prior (x_prior, P_prior), and self.z is set to None.

        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.

        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.

        H : np.array, or None
            Optionally provide H to override the measurement function for this
            one call, otherwise self.H will be used.
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = np.zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        if R is None:
            R = self.R
        elif np.isscalar(R):
            R = np.eye(self.dim_z) * R

        if H is None:
            H = self.H

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - np.dot(H, self.x)

        # common subexpression for speed
        PHT = np.dot(self.P, H.T)

        # S = HPH' + R
        # project system uncertainty into measurement space
        self.S = np.dot(H, PHT) + R
        self.SI = self.inv(self.S)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = np.dot(PHT, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + np.dot(self.K, self.y)

        # P = (I-KH)P(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P = (I-KH)P usually seen in the literature.

        I_KH = self._I - np.dot(self.K, H)
        self.P = np.dot(np.dot(I_KH, self.P), I_KH.T) + np.dot(
            np.dot(self.K, R), self.K.T
        )

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def predict_steadystate(self, u=0, B=None):
        """
        Predict state (prior) using the Kalman filter state propagation
        equations. Only x is updated, P is left unchanged. See
        update_steadstate() for a longer explanation of when to use this
        method.

        Parameters
        ----------

        u : np.array
            Optional control vector. If non-zero, it is multiplied by B
            to create the control input into the system.

        B : np.array(dim_x, dim_z), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.
        """

        if B is None:
            B = self.B

        # x = Fx + Bu
        if B is not None:
            self.x = np.dot(self.F, self.x) + np.dot(B, u)
        else:
            self.x = np.dot(self.F, self.x)

        # save prior
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def update_steadystate(self, z):
        """
        Add a new measurement (z) to the Kalman filter without recomputing
        the Kalman gain K, the state covariance P, or the system
        uncertainty S.

        You can use this for LTI systems since the Kalman gain and covariance
        converge to a fixed value. Precompute these and assign them explicitly,
        or run the Kalman filter using the normal predict()/update(0 cycle
        until they converge.

        The main advantage of this call is speed. We do significantly less
        computation, notably avoiding a costly matrix inversion.

        Use in conjunction with predict_steadystate(), otherwise P will grow
        without bound.

        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.


        Examples
        --------
        >>> cv = kinematic_kf(dim=3, order=2) # 3D const velocity filter
        >>> # let filter converge on representative data, then save k and P
        >>> for i in range(100):
        >>>     cv.predict()
        >>>     cv.update([i, i, i])
        >>> saved_k = np.copy(cv.K)
        >>> saved_P = np.copy(cv.P)

        later on:

        >>> cv = kinematic_kf(dim=3, order=2) # 3D const velocity filter
        >>> cv.K = np.copy(saved_K)
        >>> cv.P = np.copy(saved_P)
        >>> for i in range(100):
        >>>     cv.predict_steadystate()
        >>>     cv.update_steadystate([i, i, i])
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = np.zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - np.dot(self.H, self.x)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + np.dot(self.K, self.y)

        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

    def update_correlated(self, z, R=None, H=None):
        """Add a new measurement (z) to the Kalman filter assuming that
        process noise and measurement noise are correlated as defined in
        the `self.M` matrix.

        If z is None, nothing is changed.

        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.

        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.

        H : np.array,  or None
            Optionally provide H to override the measurement function for this
            one call, otherwise  self.H will be used.
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = np.zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        if R is None:
            R = self.R
        elif np.isscalar(R):
            R = np.eye(self.dim_z) * R

        # rename for readability and a tiny extra bit of speed
        if H is None:
            H = self.H

        # handle special case: if z is in form [[z]] but x is not a column
        # vector dimensions will not match
        if self.x.ndim == 1 and np.shape(z) == (1, 1):
            z = z[0]

        if np.shape(z) == ():  # is it scalar, e.g. z=3 or z=np.array(3)
            z = np.asarray([z])

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - np.dot(H, self.x)

        # common subexpression for speed
        PHT = np.dot(self.P, H.T)

        # project system uncertainty into measurement space
        self.S = np.dot(H, PHT) + np.dot(H, self.M) + np.dot(self.M.T, H.T) + R
        self.SI = self.inv(self.S)

        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = np.dot(PHT + self.M, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + np.dot(self.K, self.y)
        self.P = self.P - np.dot(self.K, np.dot(H, self.P) + self.M.T)

        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def batch_filter(
        self,
        zs,
        Fs=None,
        Qs=None,
        Hs=None,
        Rs=None,
        Bs=None,
        us=None,
        update_first=False,
        saver=None,
    ):
        """Batch processes a sequences of measurements.

         Parameters
         ----------

         zs : list-like
             list of measurements at each time step `self.dt`. Missing
             measurements must be represented by `None`.

         Fs : None, list-like, default=None
             optional value or list of values to use for the state transition
             matrix F.

             If Fs is None then self.F is used for all epochs.

             Otherwise it must contain a list-like list of F's, one for
             each epoch.  This allows you to have varying F per epoch.

         Qs : None, np.array or list-like, default=None
             optional value or list of values to use for the process error
             covariance Q.

             If Qs is None then self.Q is used for all epochs.

             Otherwise it must contain a list-like list of Q's, one for
             each epoch.  This allows you to have varying Q per epoch.

         Hs : None, np.array or list-like, default=None
             optional list of values to use for the measurement matrix H.

             If Hs is None then self.H is used for all epochs.

             If Hs contains a single matrix, then it is used as H for all
             epochs.

             Otherwise it must contain a list-like list of H's, one for
             each epoch.  This allows you to have varying H per epoch.

         Rs : None, np.array or list-like, default=None
             optional list of values to use for the measurement error
             covariance R.

             If Rs is None then self.R is used for all epochs.

             Otherwise it must contain a list-like list of R's, one for
             each epoch.  This allows you to have varying R per epoch.

         Bs : None, np.array or list-like, default=None
             optional list of values to use for the control transition matrix B.

             If Bs is None then self.B is used for all epochs.

             Otherwise it must contain a list-like list of B's, one for
             each epoch.  This allows you to have varying B per epoch.

         us : None, np.array or list-like, default=None
             optional list of values to use for the control input vector;

             If us is None then None is used for all epochs (equivalent to 0,
             or no control input).

             Otherwise it must contain a list-like list of u's, one for
             each epoch.

        update_first : bool, optional, default=False
             controls whether the order of operations is update followed by
             predict, or predict followed by update. Default is predict->update.

         saver : filterpy.common.Saver, optional
             filterpy.common.Saver object. If provided, saver.save() will be
             called after every epoch

         Returns
         -------

         means : np.array((n,dim_x,1))
             array of the state for each time step after the update. Each entry
             is an np.array. In other words `means[k,:]` is the state at step
             `k`.

         covariance : np.array((n,dim_x,dim_x))
             array of the covariances for each time step after the update.
             In other words `covariance[k,:,:]` is the covariance at step `k`.

         means_predictions : np.array((n,dim_x,1))
             array of the state for each time step after the predictions. Each
             entry is an np.array. In other words `means[k,:]` is the state at
             step `k`.

         covariance_predictions : np.array((n,dim_x,dim_x))
             array of the covariances for each time step after the prediction.
             In other words `covariance[k,:,:]` is the covariance at step `k`.

         Examples
         --------

         .. code-block:: Python

             # this example demonstrates tracking a measurement where the time
             # between measurement varies, as stored in dts. This requires
             # that F be recomputed for each epoch. The output is then smoothed
             # with an RTS smoother.

             zs = [t + random.randn()*4 for t in range (40)]
             Fs = [np.array([[1., dt], [0, 1]] for dt in dts]

             (mu, cov, _, _) = kf.batch_filter(zs, Fs=Fs)
             (xs, Ps, Ks) = kf.rts_smoother(mu, cov, Fs=Fs)
        """

        # pylint: disable=too-many-statements
        n = np.size(zs, 0)
        if Fs is None:
            Fs = [self.F] * n
        if Qs is None:
            Qs = [self.Q] * n
        if Hs is None:
            Hs = [self.H] * n
        if Rs is None:
            Rs = [self.R] * n
        if Bs is None:
            Bs = [self.B] * n
        if us is None:
            us = [0] * n

        # mean estimates from Kalman Filter
        if self.x.ndim == 1:
            means = np.zeros((n, self.dim_x))
            means_p = np.zeros((n, self.dim_x))
        else:
            means = np.zeros((n, self.dim_x, 1))
            means_p = np.zeros((n, self.dim_x, 1))

        # state covariances from Kalman Filter
        covariances = np.zeros((n, self.dim_x, self.dim_x))
        covariances_p = np.zeros((n, self.dim_x, self.dim_x))

        if update_first:
            for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):
                self.update(z, R=R, H=H)
                means[i, :] = self.x
                covariances[i, :, :] = self.P

                self.predict(u=u, B=B, F=F, Q=Q)
                means_p[i, :] = self.x
                covariances_p[i, :, :] = self.P

                if saver is not None:
                    saver.save()
        else:
            for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):
                self.predict(u=u, B=B, F=F, Q=Q)
                means_p[i, :] = self.x
                covariances_p[i, :, :] = self.P

                self.update(z, R=R, H=H)
                means[i, :] = self.x
                covariances[i, :, :] = self.P

                if saver is not None:
                    saver.save()

        return (means, covariances, means_p, covariances_p)

    def rts_smoother(self, Xs, Ps, Fs=None, Qs=None, inv=np.linalg.inv):
        """
        Runs the Rauch-Tung-Striebal Kalman smoother on a set of
        means and covariances computed by a Kalman filter. The usual input
        would come from the output of `KalmanFilter.batch_filter()`.

        Parameters
        ----------

        Xs : numpy.array
           array of the means (state variable x) of the output of a Kalman
           filter.

        Ps : numpy.array
            array of the covariances of the output of a kalman filter.

        Fs : list-like collection of numpy.array, optional
            State transition matrix of the Kalman filter at each time step.
            Optional, if not provided the filter's self.F will be used

        Qs : list-like collection of numpy.array, optional
            Process noise of the Kalman filter at each time step. Optional,
            if not provided the filter's self.Q will be used

        inv : function, default numpy.linalg.inv
            If you prefer another inverse function, such as the Moore-Penrose
            pseudo inverse, set it to that instead: kf.inv = np.linalg.pinv


        Returns
        -------

        x : numpy.ndarray
           smoothed means

        P : numpy.ndarray
           smoothed state covariances

        K : numpy.ndarray
            smoother gain at each step

        Pp : numpy.ndarray
           Predicted state covariances

        Examples
        --------

        .. code-block:: Python

            zs = [t + random.randn()*4 for t in range (40)]

            (mu, cov, _, _) = kalman.batch_filter(zs)
            (x, P, K, Pp) = rts_smoother(mu, cov, kf.F, kf.Q)

        """

        if len(Xs) != len(Ps):
            raise ValueError("length of Xs and Ps must be the same")

        n = Xs.shape[0]
        dim_x = Xs.shape[1]

        if Fs is None:
            Fs = [self.F] * n
        if Qs is None:
            Qs = [self.Q] * n

        # smoother gain
        K = np.zeros((n, dim_x, dim_x))

        x, P, Pp = Xs.copy(), Ps.copy(), Ps.copy()
        for k in range(n - 2, -1, -1):
            Pp[k] = np.dot(np.dot(Fs[k + 1], P[k]), Fs[k + 1].T) + Qs[k + 1]

            # pylint: disable=bad-whitespace
            K[k] = np.dot(np.dot(P[k], Fs[k + 1].T), inv(Pp[k]))
            x[k] += np.dot(K[k], x[k + 1] - np.dot(Fs[k + 1], x[k]))
            P[k] += np.dot(np.dot(K[k], P[k + 1] - Pp[k]), K[k].T)

        return (x, P, K, Pp)

    def get_prediction(self, u=0):
        """
        Predicts the next state of the filter and returns it without
        altering the state of the filter.

        Parameters
        ----------

        u : np.array
            optional control input

        Returns
        -------

        (x, P) : tuple
            State vector and covariance array of the prediction.
        """

        x = np.dot(self.F, self.x) + np.dot(self.B, u)
        P = self._alpha_sq * np.dot(np.dot(self.F, self.P), self.F.T) + self.Q
        return (x, P)

    def get_update(self, z=None):
        """
        Computes the new estimate based on measurement `z` and returns it
        without altering the state of the filter.

        Parameters
        ----------

        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.

        Returns
        -------

        (x, P) : tuple
            State vector and covariance array of the update.
        """

        if z is None:
            return self.x, self.P
        z = reshape_z(z, self.dim_z, self.x.ndim)

        R = self.R
        H = self.H
        P = self.P
        x = self.x

        # error (residual) between measurement and prediction
        y = z - np.dot(H, x)

        # common subexpression for speed
        PHT = np.dot(P, H.T)

        # project system uncertainty into measurement space
        S = np.dot(H, PHT) + R

        # map system uncertainty into kalman gain
        K = np.dot(PHT, self.inv(S))

        # predict new x with residual scaled by the kalman gain
        x = x + np.dot(K, y)

        # P = (I-KH)P(I-KH)' + KRK'
        I_KH = self._I - np.dot(K, H)
        P = np.dot(np.dot(I_KH, P), I_KH.T) + np.dot(np.dot(K, R), K.T)

        return x, P

    def residual_of(self, z):
        """
        Returns the residual for the given measurement (z). Does not alter
        the state of the filter.
        """
        return z - np.dot(self.H, self.x_prior)

    def measurement_of_state(self, x):
        """
        Helper function that converts a state into a measurement.

        Parameters
        ----------

        x : np.array
            kalman state vector

        Returns
        -------

        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
        """

        return np.dot(self.H, x)

    @property
    def log_likelihood(self):
        """
        log-likelihood of the last measurement.
        """
        if self._log_likelihood is None:
            self._log_likelihood = logpdf(x=self.y, cov=self.S)
        return self._log_likelihood

    @property
    def likelihood(self):
        """
        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.
        """
        if self._likelihood is None:
            self._likelihood = np.exp(self.log_likelihood)
            if self._likelihood == 0:
                self._likelihood = sys.float_info.min
        return self._likelihood

    @property
    def mahalanobis(self):
        """ "
        Mahalanobis distance of measurement. E.g. 3 means measurement
        was 3 standard deviations away from the predicted value.

        Returns
        -------
        mahalanobis : float
        """
        if self._mahalanobis is None:
            self._mahalanobis = np.sqrt(
                float(np.dot(np.dot(self.y.T, self.SI), self.y))
            )
        return self._mahalanobis

    @property
    def alpha(self):
        """
        Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading
        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon [1]_.
        """
        return self._alpha_sq**0.5

    def log_likelihood_of(self, z):
        """
        log likelihood of the measurement `z`. This should only be called
        after a call to update(). Calling after predict() will yield an
        incorrect result."""

        if z is None:
            return np.log(sys.float_info.min)
        return logpdf(z, np.dot(self.H, self.x), self.S)

    @alpha.setter
    def alpha(self, value):
        if not np.isscalar(value) or value < 1:
            raise ValueError("alpha must be a float greater than 1")

        self._alpha_sq = value**2

    def __repr__(self):
        return "\n".join(
            [
                "KalmanFilter object",
                pretty_str("dim_x", self.dim_x),
                pretty_str("dim_z", self.dim_z),
                pretty_str("dim_u", self.dim_u),
                pretty_str("x", self.x),
                pretty_str("P", self.P),
                pretty_str("x_prior", self.x_prior),
                pretty_str("P_prior", self.P_prior),
                pretty_str("x_post", self.x_post),
                pretty_str("P_post", self.P_post),
                pretty_str("F", self.F),
                pretty_str("Q", self.Q),
                pretty_str("R", self.R),
                pretty_str("H", self.H),
                pretty_str("K", self.K),
                pretty_str("y", self.y),
                pretty_str("S", self.S),
                pretty_str("SI", self.SI),
                pretty_str("M", self.M),
                pretty_str("B", self.B),
                pretty_str("z", self.z),
                pretty_str("log-likelihood", self.log_likelihood),
                pretty_str("likelihood", self.likelihood),
                pretty_str("mahalanobis", self.mahalanobis),
                pretty_str("alpha", self.alpha),
                pretty_str("inv", self.inv),
            ]
        )

    def test_matrix_dimensions(self, z=None, H=None, R=None, F=None, Q=None):
        """
        Performs a series of asserts to check that the size of everything
        is what it should be. This can help you debug problems in your design.

        If you pass in H, R, F, Q those will be used instead of this object's
        value for those matrices.

        Testing `z` (the measurement) is problamatic. x is a vector, and can be
        implemented as either a 1D array or as a nx1 column vector. Thus Hx
        can be of different shapes. Then, if Hx is a single value, it can
        be either a 1D array or 2D vector. If either is true, z can reasonably
        be a scalar (either '3' or np.array('3') are scalars under this
        definition), a 1D, 1 element array, or a 2D, 1 element array. You are
        allowed to pass in any combination that works.
        """

        if H is None:
            H = self.H
        if R is None:
            R = self.R
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        x = self.x
        P = self.P

        assert (
            x.ndim == 1 or x.ndim == 2
        ), "x must have one or two dimensions, but has {}".format(x.ndim)

        if x.ndim == 1:
            assert (
                x.shape[0] == self.dim_x
            ), "Shape of x must be ({},{}), but is {}".format(self.dim_x, 1, x.shape)
        else:
            assert x.shape == (
                self.dim_x,
                1,
            ), "Shape of x must be ({},{}), but is {}".format(self.dim_x, 1, x.shape)

        assert P.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of P must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, P.shape
        )

        assert Q.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of P must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, P.shape
        )

        assert F.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of F must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, F.shape
        )

        assert np.ndim(H) == 2, "Shape of H must be (dim_z, {}), but is {}".format(
            P.shape[0], np.shape(H)
        )

        assert (
            H.shape[1] == P.shape[0]
        ), "Shape of H must be (dim_z, {}), but is {}".format(P.shape[0], H.shape)

        # shape of R must be the same as HPH'
        hph_shape = (H.shape[0], H.shape[0])
        r_shape = np.shape(R)

        if H.shape[0] == 1:
            # r can be scalar, 1D, or 2D in this case
            assert (
                r_shape == () or r_shape == (1,) or r_shape == (1, 1)
            ), "R must be scalar or one element array, but is shaped {}".format(r_shape)
        else:
            assert r_shape == hph_shape, "shape of R should be {} but it is {}".format(
                hph_shape, r_shape
            )

        if z is not None:
            z_shape = np.shape(z)
        else:
            z_shape = (self.dim_z, 1)

        # H@x must have shape of z
        Hx = np.dot(H, x)

        if z_shape == ():  # scalar or np.array(scalar)
            assert Hx.ndim == 1 or np.shape(Hx) == (
                1,
                1,
            ), "shape of z should be {}, not {} for the given H".format(
                np.shape(Hx), z_shape
            )

        elif np.shape(Hx) == (1,):
            assert z_shape[0] == 1, "Shape of z must be {} for the given H".format(
                np.shape(Hx)
            )

        else:
            assert z_shape == np.shape(Hx) or (
                len(z_shape) == 1 and np.shape(Hx) == (z_shape[0], 1)
            ), "shape of z should be {}, not {} for the given H".format(
                np.shape(Hx), z_shape
            )

        if np.ndim(Hx) > 1 and np.shape(Hx) != (1, 1):
            assert (
                np.shape(Hx) == z_shape
            ), "shape of z should be {} for the given H, but it is {}".format(
                np.shape(Hx), z_shape
            )


def update(x, P, z, R, H=None, return_all=False):
    """
    Add a new measurement (z) to the Kalman filter. If z is None, nothing
    is changed.

    This can handle either the multidimensional or unidimensional case. If
    all parameters are floats instead of arrays the filter will still work,
    and return floats for x, P as the result.

    update(1, 2, 1, 1, 1)  # univariate
    update(x, P, 1



    Parameters
    ----------

    x : numpy.array(dim_x, 1), or float
        State estimate vector

    P : numpy.array(dim_x, dim_x), or float
        Covariance matrix

    z : (dim_z, 1): array_like
        measurement for this update. z can be a scalar if dim_z is 1,
        otherwise it must be convertible to a column vector.

    R : numpy.array(dim_z, dim_z), or float
        Measurement noise matrix

    H : numpy.array(dim_x, dim_x), or float, optional
        Measurement function. If not provided, a value of 1 is assumed.

    return_all : bool, default False
        If true, y, K, S, and log_likelihood are returned, otherwise
        only x and P are returned.

    Returns
    -------

    x : numpy.array
        Posterior state estimate vector

    P : numpy.array
        Posterior covariance matrix

    y : numpy.array or scalar
        Residua. Difference between measurement and state in measurement space

    K : numpy.array
        Kalman gain

    S : numpy.array
        System uncertainty in measurement space

    log_likelihood : float
        log likelihood of the measurement
    """

    # pylint: disable=bare-except

    if z is None:
        if return_all:
            return x, P, None, None, None, None
        return x, P

    if H is None:
        H = np.array([1])

    if np.isscalar(H):
        H = np.array([H])

    Hx = np.atleast_1d(np.dot(H, x))
    z = reshape_z(z, Hx.shape[0], x.ndim)

    # error (residual) between measurement and prediction
    y = z - Hx

    # project system uncertainty into measurement space
    S = np.dot(np.dot(H, P), H.T) + R

    # map system uncertainty into kalman gain
    try:
        K = np.dot(np.dot(P, H.T), np.linalg.inv(S))
    except Exception:
        # can't invert a 1D array, annoyingly
        K = np.dot(np.dot(P, H.T), 1.0 / S)

    # predict new x with residual scaled by the kalman gain
    x = x + np.dot(K, y)

    # P = (I-KH)P(I-KH)' + KRK'
    KH = np.dot(K, H)

    try:
        I_KH = np.eye(KH.shape[0]) - KH
    except Exception:
        I_KH = np.array([1 - KH])
    P = np.dot(np.dot(I_KH, P), I_KH.T) + np.dot(np.dot(K, R), K.T)

    if return_all:
        # compute log likelihood
        log_likelihood = logpdf(z, np.dot(H, x), S)
        return x, P, y, K, S, log_likelihood
    return x, P


def update_steadystate(x, z, K, H=None):
    """
    Add a new measurement (z) to the Kalman filter. If z is None, nothing
    is changed.


    Parameters
    ----------

    x : numpy.array(dim_x, 1), or float
        State estimate vector


    z : (dim_z, 1): array_like
        measurement for this update. z can be a scalar if dim_z is 1,
        otherwise it must be convertible to a column vector.

    K : numpy.array, or float
        Kalman gain matrix

    H : numpy.array(dim_x, dim_x), or float, optional
        Measurement function. If not provided, a value of 1 is assumed.

    Returns
    -------

    x : numpy.array
        Posterior state estimate vector

    Examples
    --------

    This can handle either the multidimensional or unidimensional case. If
    all parameters are floats instead of arrays the filter will still work,
    and return floats for x, P as the result.

    >>> update_steadystate(1, 2, 1)  # univariate
    >>> update_steadystate(x, P, z, H)
    """

    if z is None:
        return x

    if H is None:
        H = np.array([1])

    if np.isscalar(H):
        H = np.array([H])

    Hx = np.atleast_1d(np.dot(H, x))
    z = reshape_z(z, Hx.shape[0], x.ndim)

    # error (residual) between measurement and prediction
    y = z - Hx

    # estimate new x with residual scaled by the kalman gain
    return x + np.dot(K, y)


def predict(x, P, F=1, Q=0, u=0, B=1, alpha=1.0):
    """
    Predict next state (prior) using the Kalman filter state propagation
    equations.

    Parameters
    ----------

    x : numpy.array
        State estimate vector

    P : numpy.array
        Covariance matrix

    F : numpy.array()
        State Transition matrix

    Q : numpy.array, Optional
        Process noise matrix


    u : numpy.array, Optional, default 0.
        Control vector. If non-zero, it is multiplied by B
        to create the control input into the system.

    B : numpy.array, optional, default 0.
        Control transition matrix.

    alpha : float, Optional, default=1.0
        Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading
        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon

    Returns
    -------

    x : numpy.array
        Prior state estimate vector

    P : numpy.array
        Prior covariance matrix
    """

    if np.isscalar(F):
        F = np.array(F)
    x = np.dot(F, x) + np.dot(B, u)
    P = (alpha * alpha) * np.dot(np.dot(F, P), F.T) + Q

    return x, P


def predict_steadystate(x, F=1, u=0, B=1):
    """
    Predict next state (prior) using the Kalman filter state propagation
    equations. This steady state form only computes x, assuming that the
    covariance is constant.

    Parameters
    ----------

    x : numpy.array
        State estimate vector

    P : numpy.array
        Covariance matrix

    F : numpy.array()
        State Transition matrix

    u : numpy.array, Optional, default 0.
        Control vector. If non-zero, it is multiplied by B
        to create the control input into the system.

    B : numpy.array, optional, default 0.
        Control transition matrix.

    Returns
    -------

    x : numpy.array
        Prior state estimate vector
    """

    if np.isscalar(F):
        F = np.array(F)
    x = np.dot(F, x) + np.dot(B, u)

    return x


def batch_filter(
    x, P, zs, Fs, Qs, Hs, Rs, Bs=None, us=None, update_first=False, saver=None
):
    """
    Batch processes a sequences of measurements.

    Parameters
    ----------

    zs : list-like
        list of measurements at each time step. Missing measurements must be
        represented by None.

    Fs : list-like
        list of values to use for the state transition matrix matrix.

    Qs : list-like
        list of values to use for the process error
        covariance.

    Hs : list-like
        list of values to use for the measurement matrix.

    Rs : list-like
        list of values to use for the measurement error
        covariance.

    Bs : list-like, optional
        list of values to use for the control transition matrix;
        a value of None in any position will cause the filter
        to use `self.B` for that time step.

    us : list-like, optional
        list of values to use for the control input vector;
        a value of None in any position will cause the filter to use
        0 for that time step.

    update_first : bool, optional
        controls whether the order of operations is update followed by
        predict, or predict followed by update. Default is predict->update.

        saver : filterpy.common.Saver, optional
            filterpy.common.Saver object. If provided, saver.save() will be
            called after every epoch

    Returns
    -------

    means : np.array((n,dim_x,1))
        array of the state for each time step after the update. Each entry
        is an np.array. In other words `means[k,:]` is the state at step
        `k`.

    covariance : np.array((n,dim_x,dim_x))
        array of the covariances for each time step after the update.
        In other words `covariance[k,:,:]` is the covariance at step `k`.

    means_predictions : np.array((n,dim_x,1))
        array of the state for each time step after the predictions. Each
        entry is an np.array. In other words `means[k,:]` is the state at
        step `k`.

    covariance_predictions : np.array((n,dim_x,dim_x))
        array of the covariances for each time step after the prediction.
        In other words `covariance[k,:,:]` is the covariance at step `k`.

    Examples
    --------

    .. code-block:: Python

        zs = [t + random.randn()*4 for t in range (40)]
        Fs = [kf.F for t in range (40)]
        Hs = [kf.H for t in range (40)]

        (mu, cov, _, _) = kf.batch_filter(zs, Rs=R_list, Fs=Fs, Hs=Hs, Qs=None,
                                          Bs=None, us=None, update_first=False)
        (xs, Ps, Ks) = kf.rts_smoother(mu, cov, Fs=Fs, Qs=None)

    """

    n = np.size(zs, 0)
    dim_x = x.shape[0]

    # mean estimates from Kalman Filter
    if x.ndim == 1:
        means = np.zeros((n, dim_x))
        means_p = np.zeros((n, dim_x))
    else:
        means = np.zeros((n, dim_x, 1))
        means_p = np.zeros((n, dim_x, 1))

    # state covariances from Kalman Filter
    covariances = np.zeros((n, dim_x, dim_x))
    covariances_p = np.zeros((n, dim_x, dim_x))

    if us is None:
        us = [0.0] * n
        Bs = [0.0] * n

    if update_first:
        for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):
            x, P = update(x, P, z, R=R, H=H)
            means[i, :] = x
            covariances[i, :, :] = P

            x, P = predict(x, P, u=u, B=B, F=F, Q=Q)
            means_p[i, :] = x
            covariances_p[i, :, :] = P
            if saver is not None:
                saver.save()
    else:
        for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):
            x, P = predict(x, P, u=u, B=B, F=F, Q=Q)
            means_p[i, :] = x
            covariances_p[i, :, :] = P

            x, P = update(x, P, z, R=R, H=H)
            means[i, :] = x
            covariances[i, :, :] = P
            if saver is not None:
                saver.save()

    return (means, covariances, means_p, covariances_p)


def rts_smoother(Xs, Ps, Fs, Qs):
    """
    Runs the Rauch-Tung-Striebal Kalman smoother on a set of
    means and covariances computed by a Kalman filter. The usual input
    would come from the output of `KalmanFilter.batch_filter()`.

    Parameters
    ----------

    Xs : numpy.array
       array of the means (state variable x) of the output of a Kalman
       filter.

    Ps : numpy.array
        array of the covariances of the output of a kalman filter.

    Fs : list-like collection of numpy.array
        State transition matrix of the Kalman filter at each time step.

    Qs : list-like collection of numpy.array, optional
        Process noise of the Kalman filter at each time step.

    Returns
    -------

    x : numpy.ndarray
       smoothed means

    P : numpy.ndarray
       smoothed state covariances

    K : numpy.ndarray
        smoother gain at each step

    pP : numpy.ndarray
       predicted state covariances

    Examples
    --------

    .. code-block:: Python

        zs = [t + random.randn()*4 for t in range (40)]

        (mu, cov, _, _) = kalman.batch_filter(zs)
        (x, P, K, pP) = rts_smoother(mu, cov, kf.F, kf.Q)
    """

    if len(Xs) != len(Ps):
        raise ValueError("length of Xs and Ps must be the same")

    n = Xs.shape[0]
    dim_x = Xs.shape[1]

    # smoother gain
    K = np.zeros((n, dim_x, dim_x))
    x, P, pP = Xs.copy(), Ps.copy(), Ps.copy()

    for k in range(n - 2, -1, -1):
        pP[k] = np.dot(np.dot(Fs[k], P[k]), Fs[k].T) + Qs[k]

        # pylint: disable=bad-whitespace
        K[k] = np.dot(np.dot(P[k], Fs[k].T), np.linalg.inv(pP[k]))
        x[k] += np.dot(K[k], x[k + 1] - np.dot(Fs[k], x[k]))
        P[k] += np.dot(np.dot(K[k], P[k + 1] - pP[k]), K[k].T)

    return (x, P, K, pP)


def reshape_z(z, dim_z, ndim):
    """ensure z is a (dim_z, 1) shaped vector"""

    z = np.atleast_2d(z)
    if z.shape[1] == dim_z:
        z = z.T

    if z.shape != (dim_z, 1):
        raise ValueError("z must be convertible to shape ({}, 1)".format(dim_z))

    if ndim == 1:
        z = z[:, 0]

    if ndim == 0:
        z = z[0, 0]

    return z


def pretty_str(label, arr):
    """
    Generates a pretty printed NumPy array with an assignment. Optionally
    transposes column vectors so they are drawn on one line. Strictly speaking
    arr can be any time convertible by `str(arr)`, but the output may not
    be what you want if the type of the variable is not a scalar or an
    ndarray.

    Examples
    --------
    >>> pprint('cov', np.array([[4., .1], [.1, 5]]))
    cov = [[4.  0.1]
           [0.1 5. ]]

    >>> print(pretty_str('x', np.array([[1], [2], [3]])))
    x = [[1 2 3]].T
    """

    def is_col(a):
        """return true if a is a column vector"""
        try:
            return a.shape[0] > 1 and a.shape[1] == 1
        except (AttributeError, IndexError):
            return False

    if label is None:
        label = ""

    if label:
        label += " = "

    if is_col(arr):
        return label + str(arr.T).replace("\n", "") + ".T"

    rows = str(arr).split("\n")
    if not rows:
        return ""

    s = label + rows[0]
    pad = " " * len(label)
    for line in rows[1:]:
        s = s + "\n" + pad + line

    return s


def logpdf(x, mean=None, cov=1, allow_singular=True):
    """
    Computes the log of the probability density function of the normal
    N(mean, cov) for the data x. The normal may be univariate or multivariate.

    Wrapper for older versions of scipy.multivariate_normal.logpdf which
    don't support support the allow_singular keyword prior to verion 0.15.0.

    If it is not supported, and cov is singular or not PSD you may get
    an exception.

    `x` and `mean` may be column vectors, row vectors, or lists.
    """

    if mean is not None:
        flat_mean = np.asarray(mean).flatten()
    else:
        flat_mean = None

    flat_x = np.asarray(x).flatten()

    return multivariate_normal.logpdf(flat_x, flat_mean, cov, allow_singular)



"""This module contains the class to process a given batch of data."""

from time import time
from typing import TYPE_CHECKING
import numpy as np

from py_neuromodulation import logger
from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation.features import FeatureProcessors
from py_neuromodulation.utils import io
from py_neuromodulation.stream.settings import NMSettings

if TYPE_CHECKING:
    from py_neuromodulation.processing.projection import Projection
    import pandas as pd


class DataProcessor:
    def __init__(
        self,
        sfreq: float,
        settings: NMSettings | _PathLike,
        channels: "pd.DataFrame | _PathLike",
        coord_names: list | None = None,
        coord_list: list | None = None,
        line_noise: float | None = None,
        path_grids: _PathLike | None = None,
        verbose: bool = True,
    ) -> None:
        from py_neuromodulation.processing import DataPreprocessor

        """Initialize run class.

        Parameters
        ----------
        settings : settings.NMSettings object
        channels : pd.DataFrame | _PathLike
            Initialized pd.DataFrame with channel specific information.
            The path to a channels.csv can be also passed.
        coord_names : list | None
            list of coordinate names
        coord_list : list | None
            list of list of 3D coordinates
        path_grids : _PathLike | None
            List to grid_cortex.tsv and grid_subcortex.tsv for grid point projection
        verbose : boolean
            if True, log signal processed and computation time
        """

        self.settings = NMSettings.load(settings)
        self.channels = io.load_channels(channels)

        self.sfreq_features: float = self.settings.sampling_rate_features_hz
        self._sfreq_raw_orig: float = sfreq
        self.sfreq_raw: float = sfreq // 1
        self.line_noise: float | None = line_noise
        self.path_grids: _PathLike | None = path_grids
        self.non_psd_indices: np.ndarray | None = None
        self.verbose: bool = verbose

        self.features_previous = None

        (self.ch_names_used, _, self.feature_idx, _) = self._get_ch_info()

        self.preprocessors = DataPreprocessor(
            settings=self.settings,
            channels=self.channels,
            sfreq=self.sfreq_raw,
            line_noise=self.line_noise,
        )

        if self.settings.postprocessing.feature_normalization:
            from py_neuromodulation.processing.normalization import FeatureNormalizer

            self.feature_normalizer = FeatureNormalizer(self.settings)

        self.features = FeatureProcessors(
            settings=self.settings,
            ch_names=self.ch_names_used,
            sfreq=self.sfreq_raw,
        )

        if coord_list is not None and coord_names is not None:
            self.coords = self._set_coords(
                coord_names=coord_names, coord_list=coord_list
            )

        self.projection = self._get_projection(self.settings, self.channels)

        self.cnt_samples = 0

    @staticmethod
    def _add_coordinates(coord_names: list[str], coord_list: list) -> dict:
        """Write cortical and subcortical coordinate information in joint dictionary

        Parameters
        ----------
        coord_names : list[str]
            list of coordinate names
        coord_list : list
            list of list of 3D coordinates

        Returns
        -------
        dict with (sub)cortex_left and (sub)cortex_right ch_names and positions
        """

        def is_left_coord(val: float, coord_region: str) -> bool:
            if coord_region.split("_")[1] == "left":
                return val < 0
            return val > 0

        coords: dict[str, dict[str, list | np.ndarray]] = {}

        for coord_region in [
            coord_loc + "_" + lat
            for coord_loc in ["cortex", "subcortex"]
            for lat in ["left", "right"]
        ]:
            coords[coord_region] = {}

            ch_type = "ECOG" if "cortex" == coord_region.split("_")[0] else "LFP"

            coords[coord_region]["ch_names"] = [
                coord_name
                for coord_name, ch in zip(coord_names, coord_list)
                if is_left_coord(ch[0], coord_region) and (ch_type in coord_name)
            ]

            # multiply by 1000 to get m instead of mm
            positions = []
            for coord, coord_name in zip(coord_list, coord_names):
                if is_left_coord(coord[0], coord_region) and (ch_type in coord_name):
                    positions.append(coord)
            coords[coord_region]["positions"] = (
                np.array(positions, dtype=np.float64) * 1000
            )

        return coords

    def _get_ch_info(
        self,
    ) -> tuple[list[str], list[str], list[int], np.ndarray]:
        """Get used feature and label info from channels"""
        channels = self.channels
        ch_names_used = channels[channels["used"] == 1]["new_name"].tolist()
        ch_types_used = channels[channels["used"] == 1]["type"].tolist()

        # used channels for feature estimation
        feature_idx = np.where(channels["used"] & ~channels["target"])[0].tolist()

        # If multiple targets exist, select only the first
        label_idx = np.where(channels["target"] == 1)[0]

        return ch_names_used, ch_types_used, feature_idx, label_idx

    @staticmethod
    def _get_grids(
        settings: "NMSettings",
        path_grids: _PathLike | None,
    ) -> "tuple[pd.DataFrame | None, pd.DataFrame | None]":
        """Read settings specified grids

        Parameters
        ----------
        settings : settings.NMSettings object
        path_grids : _PathLike | str

        Returns
        -------
        Tuple
            grid_cortex, grid_subcortex,
            might be None if not specified in settings
        """
        if settings.postprocessing.project_cortex:
            grid_cortex = io.read_grid(path_grids, "cortex")
        else:
            grid_cortex = None
        if settings.postprocessing.project_subcortex:
            grid_subcortex = io.read_grid(path_grids, "subcortex")
        else:
            grid_subcortex = None
        return grid_cortex, grid_subcortex

    def _get_projection(
        self, settings: "NMSettings", channels: "pd.DataFrame"
    ) -> "Projection | None":
        from py_neuromodulation.processing.projection import Projection

        """Return projection of used coordinated and grids"""

        if not any(
            (
                settings.postprocessing.project_cortex,
                settings.postprocessing.project_subcortex,
            )
        ):
            return None

        grid_cortex, grid_subcortex = self._get_grids(self.settings, self.path_grids)
        projection = Projection(
            settings=settings,
            grid_cortex=grid_cortex,
            grid_subcortex=grid_subcortex,
            coords=self.coords,
            channels=channels,
            plot_projection=False,
        )
        return projection

    def _set_coords(
        self, coord_names: list[str] | None, coord_list: list | None
    ) -> dict:
        if not any(
            (
                self.settings.postprocessing.project_cortex,
                self.settings.postprocessing.project_subcortex,
            )
        ):
            return {}

        if any((coord_list is None, coord_names is None)):
            raise ValueError(
                "No coordinates could be loaded. Please provide coord_list and"
                f" coord_names. Got: {coord_list=}, {coord_names=}."
            )

        return self._add_coordinates(
            coord_names=coord_names,
            coord_list=coord_list,  # type: ignore # None case handled above
        )

    def process(self, data: np.ndarray) -> dict[str, float]:
        """Given a new data batch, calculate and return features.

        Parameters
        ----------
        data : np.ndarray
            Current batch of raw data

        Returns
        -------
        pandas Series
            Features calculated from current data
        """
        start_time = time()

        nan_channels = np.isnan(data).any(axis=1)

        data = np.nan_to_num(data)[self.feature_idx, :]

        data = self.preprocessors.process_data(data)

        # calculate features
        features_dict = self.features.estimate_features(data)

        # normalize features
        if self.settings.postprocessing.feature_normalization:
            if not self.settings.feature_normalization_settings.normalize_psd:
                if self.non_psd_indices is None:
                    self.non_psd_indices = [
                        idx
                        for idx, key in enumerate(features_dict.keys())
                        if "psd" not in key
                    ]
                    self.psd_indices = list(set(range(len(features_dict))) - set(
                        self.non_psd_indices
                    ))
                feature_values = np.fromiter(features_dict.values(), dtype=np.float64)
                normed_features_non_psd = self.feature_normalizer.process(
                    feature_values[self.non_psd_indices]
                )

                # combine values in new array
                normed_features = np.empty((feature_values.shape[0]))
                normed_features[self.non_psd_indices] = normed_features_non_psd
                normed_features[self.psd_indices] = feature_values[self.psd_indices]
            else:    
                normed_features = self.feature_normalizer.process(
                    np.fromiter(features_dict.values(), dtype=np.float64)
                )
            features_dict = {
                key: normed_features[idx]
                for idx, key in enumerate(features_dict.keys())
            }

        # project features to grid
        if self.projection:
            self.projection.project_features(features_dict)

        # check for all features, where the channel had a NaN, that the feature is also put to NaN
        if nan_channels.sum() > 0:
            # TONI: no need to do this if we store both old and new names for the channels
            new_nan_channels = []
            for ch in list(np.array(self.ch_names_used)[nan_channels]):
                for key in features_dict.keys():
                    if ch in key:
                        new_nan_channels.append(key)

            for ch in new_nan_channels:
                features_dict[ch] = np.nan

        if self.verbose:
            logger.info("Last batch took: %.3f seconds to process", time() - start_time)

        return features_dict

    def save_sidecar(
        self,
        out_dir: _PathLike,
        prefix: str = "",
        additional_args: dict | None = None,
    ) -> None:
        """Save sidecar incuding fs, coords, sess_right to out_dir."""

        sidecar: dict = {
            "original_fs": self._sfreq_raw_orig,
            "final_fs": self.sfreq_raw,
            "sfreq": self.sfreq_features,
        }
        if self.projection:
            sidecar["coords"] = self.projection.coords
            if self.settings.postprocessing.project_cortex:
                sidecar["grid_cortex"] = self.projection.grid_cortex
                sidecar["proj_matrix_cortex"] = self.projection.proj_matrix_cortex
            if self.settings.postprocessing.project_subcortex:
                sidecar["grid_subcortex"] = self.projection.grid_subcortex
                sidecar["proj_matrix_subcortex"] = self.projection.proj_matrix_subcortex
        if additional_args is not None:
            sidecar = sidecar | additional_args

        io.save_sidecar(sidecar, out_dir, prefix)

    def save_settings(self, out_dir: _PathLike, prefix: str = "") -> None:
        self.settings.save(out_dir, prefix)

    def save_channels(self, out_dir: _PathLike, prefix: str) -> None:
        io.save_channels(self.channels, out_dir, prefix)

    def save_features(
        self,
        feature_arr: "pd.DataFrame",
        out_dir: _PathLike = "",
        prefix: str = "",
    ) -> None:
        io.save_features(feature_arr, out_dir, prefix)



from .generator import RawDataGenerator
from .mnelsl_player import LSLOfflinePlayer
from .stream import Stream
from .settings import NMSettings

try:
    from .mnelsl_stream import LSLStream
except Exception as e:
    print(f"A RuntimeError occurred: {e}")



from typing import TYPE_CHECKING, Type
from py_neuromodulation.utils.types import PREPROCESSOR_NAME, NMPreprocessor

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
    from py_neuromodulation.stream.settings import NMSettings

PREPROCESSOR_DICT: dict[PREPROCESSOR_NAME, str] = {
    "preprocessing_filter": "PreprocessingFilter",
    "notch_filter": "NotchFilter",
    "raw_resampling": "Resampler",
    "re_referencing": "ReReferencer",
    "raw_normalization": "RawNormalizer",
}


class DataPreprocessor:
    "Class for initializing and holding data preprocessing classes"

    def __init__(
        self,
        settings: "NMSettings",
        channels: "pd.DataFrame",
        sfreq: float,
        line_noise: float | None = None,
    ) -> None:
        from importlib import import_module
        from inspect import getfullargspec

        possible_arguments = {
            "sfreq": sfreq,
            "settings": settings,
            "channels": channels,
            "line_noise": line_noise,
        }

        for preprocessing_method in settings.preprocessing:
            if preprocessing_method not in PREPROCESSOR_DICT.keys():
                raise ValueError(
                    f"Invalid preprocessing method '{preprocessing_method}'. Must be one of {PREPROCESSOR_DICT.keys()}"
                )

        # Get needed preprocessor classes from settings
        preprocessor_classes: dict[str, Type[NMPreprocessor]] = {
            preprocessor_name: getattr(
                import_module("py_neuromodulation.processing"), class_name
            )
            for preprocessor_name, class_name in PREPROCESSOR_DICT.items()
            if preprocessor_name in settings.preprocessing
        }

        # Function to instantiate preprocessor with settings
        def instantiate_preprocessor(
            preprocessor_class: Type[NMPreprocessor], preprocessor_name: str
        ) -> NMPreprocessor:
            settings_str = f"{preprocessor_name}_settings"
            # Filter out arguments that are not in the preprocessor's __init__ method
            args = {
                arg: possible_arguments[arg]
                for arg in getfullargspec(preprocessor_class).args
                if arg in possible_arguments
            }
            # Retrieve more possible arguments from settings
            args |= getattr(settings, settings_str, {})
            # Pass arguments to preprocessor class and return instance
            return preprocessor_class(**args)

        self.preprocessors: list[NMPreprocessor] = [
            instantiate_preprocessor(preprocessor_class, preprocessor_name)
            for preprocessor_name, preprocessor_class in preprocessor_classes.items()
        ]

    def process_data(self, data: "np.ndarray") -> "np.ndarray":
        """
        Args:
            data (np.ndarray): shape: (n_channels, n_samples)

        Returns:
            np.ndarray: shape: (n_channels, n_samples)
        """
        for preprocessor in self.preprocessors:
            data = preprocessor.process(data)
        return data



import numpy as np
import mne
from pathlib import Path
import multiprocessing as mp
import atexit
import time
import signal

from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation.utils.io import read_BIDS_data
from py_neuromodulation import logger


class LSLOfflinePlayer:
    _instances: set["LSLOfflinePlayer"] = set()  # Keep track of initialized players
    _atexit_registered: bool = False  # Flag to register atexit

    def __init__(
        self,
        stream_name: str | None = "lsl_offline_player",
        f_name: str | _PathLike = None,
        raw: mne.io.Raw | None = None,
        sfreq: int | float | None = None,
        data: np.ndarray | None = None,
        ch_type: str | None = "dbs",
        chunk_size: int = 10,
        n_repeat: int = 1,
    ) -> None:
        """Initialization of MNE-LSL offline player.
        Either a filename (PathLike) is provided,
        or data and sampling frequency to initialize an example mock-up stream.


        Parameters
        ----------
        stream_name : str, optional
            LSL stream name, by default "example_stream"
        f_name : str | None, optional
            file name used for streaming, by default None
        sfreq : int | float | None, optional
            sampling rate, by default None
        data : np.ndarray | None, optional
            data used for streaming, by default None
        ch_type: str | None, optional
            channel type to select for streaming, by default "dbs"

        Raises
        ------
        ValueError
            _description_
        """
        self.sfreq = sfreq
        self.stream_name = stream_name
        self.chunk_size = chunk_size
        self.n_repeat = n_repeat

        if f_name:
            (self._path_raw, data, sfreq, line_noise, coord_list, coord_names) = (
                read_BIDS_data(f_name)
            )
        elif raw:
            self._path_raw = raw
        elif sfreq and data:
            info = mne.create_info(
                ch_names=[f"ch{i}" for i in range(data.shape[0])],
                ch_types=[ch_type for _ in range(data.shape[0])],
                sfreq=sfreq,
            )
            raw = mne.io.RawArray(data, info)
            self._path_raw = Path.cwd() / "temp_raw.fif"
            raw.save(self._path_raw, overwrite=True)
        else:
            error_msg = "Either f_name or raw or sfreq and data must be provided."
            logger.critical(error_msg)
            raise ValueError(error_msg)

        # Flags to control the player subprocess
        self._streaming_complete = mp.Event()
        self._player_process = None
        self._stop_flag = mp.Event()

        LSLOfflinePlayer._instances.add(self)  # Register instancwe
        if LSLOfflinePlayer._atexit_registered:
            atexit.register(LSLOfflinePlayer._stop_all_players)
            LSLOfflinePlayer._atexit_registered = True

    def start_player(
        self,
        chunk_size: int | None = None,
        n_repeat: int | None = None,
        block: bool = False,
    ):
        """Start MNE-LSL Player

        Parameters
        ----------
        chunk_size : int, optional
            Number of samples to stream at once, by default 10
        n_repeat : int, optional
            Number of times to repeat the stream, by default 1
        block : bool, optional
            If True, block until streaming is complete, by default False
        """

        if chunk_size:
            self.chunk_size = chunk_size
        if n_repeat:
            self.n_repeat = n_repeat

        self._stop_flag.clear()
        self._streaming_complete.clear()

        self._player_process = mp.Process(
            target=self._run_player,
            args=(
                self.chunk_size,
                self.n_repeat,
                self._stop_flag,
                self._streaming_complete,
            ),
        )
        self._player_process.start()

        if block:
            try:
                self.wait_for_completion()
            except KeyboardInterrupt:
                logger.info("\nKeyboard interrupt received. Stopping the player...")
                self.stop_player()

    def _run_player(self, chunk_size, n_repeat, stop_flag, streaming_complete):
        from mne_lsl.player import PlayerLSL

        signal.signal(signal.SIGINT, lambda: stop_flag.set())

        player = PlayerLSL(
            self._path_raw,
            name=self.stream_name,
            chunk_size=chunk_size,
            n_repeat=n_repeat,
        )
        player = player.start()

        try:
            while not stop_flag.is_set() and not player._end_streaming:
                time.sleep(0.1)
        finally:
            try:
                player.stop()
            except RuntimeError:
                # player already stopped
                pass
            streaming_complete.set()

    def wait_for_completion(self):
        """Block until streaming is complete"""
        while self._player_process and self._player_process.is_alive():
            try:
                self._streaming_complete.wait(timeout=1.0)
                if self._streaming_complete.is_set():
                    break
            except KeyboardInterrupt:
                logger.info("\nKeyboard interrupt received. Stopping the player...")
                self.stop_player()
                break

    def stop_player(self):
        """Stop MNE-LSL Player"""
        if self._player_process and self._player_process.is_alive():
            self._stop_flag.set()
            self._player_process.join(timeout=5)
            if self._player_process.is_alive():
                self._player_process.terminate()
                self._player_process.join(timeout=1)
            if self._player_process.is_alive():
                self._player_process.kill()
            self._player_process = None

        logger.info(f"Player stopped: {self.stream_name}")
        LSLOfflinePlayer._instances.discard(self)

    @classmethod
    def _stop_all_players(cls):
        """Stop all player instances (used for atexit)"""
        for player in cls._instances:
            player.stop_player()

    # Enable use as a context manager
    def __enter__(self):
        self.start_player()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_player()



from sklearn import model_selection
from sklearn.linear_model import LinearRegression
from sklearn.base import clone
from sklearn.metrics import r2_score

import pandas as pd
import numpy as np
from copy import deepcopy
from pathlib import Path, PurePath
import pickle

from py_neuromodulation import logger
from py_neuromodulation.utils.types import _PathLike

from typing import Callable


class RealTimeDecoder:
    def __init__(self, model_path: _PathLike):
        self.model_path = Path(model_path)
        if not self.model_path.exists():
            raise FileNotFoundError(f"Model file {self.model_path} not found")
        if not self.model_path.is_file():
            raise IsADirectoryError(f"Model file {self.model_path} is a directory")

        if self.model_path.suffix == ".skops":
            from skops import io as skops_io

            self.model = skops_io.load(self.model_path)
        else:
            return NotImplementedError("Only skops models are supported")

    def predict(
        self,
        feature_dict: dict,
        channel: str | None = None,
        fft_bands_only: bool = True,
    ) -> dict:
        try:
            if channel is not None:
                features_ch = {
                    f: feature_dict[f]
                    for f in feature_dict.keys()
                    if f.startswith(channel)
                }
                if fft_bands_only is True:
                    features_ch_fft = {
                        f: features_ch[f]
                        for f in features_ch.keys()
                        if "fft" in f and "psd" not in f
                    }
                    out = self.model.predict_proba(
                        np.array(list(features_ch_fft.values())).reshape(1, -1)
                    )
                else:
                    out = self.model.predict_proba(features_ch)
            else:
                out = self.model.predict(feature_dict)
            for decode_output_idx in range(out.shape[1]):
                feature_dict[f"decode_{decode_output_idx}"] = np.squeeze(out)[
                    decode_output_idx
                ]
            logger.debug(f"Decoded values: {out}")
            return feature_dict
        except Exception as e:
            logger.error(f"Error in decoding: {e}")
        return feature_dict


class CV_res:
    def __init__(
        self,
        get_movement_detection_rate: bool = False,
        RUN_BAY_OPT: bool = False,
        mrmr_select: bool = False,
        model_save: bool = False,
    ) -> None:
        self.score_train: list = []
        self.score_test: list = []
        self.y_test: list = []
        self.y_train: list = []
        self.y_test_pr: list = []
        self.y_train_pr: list = []
        self.X_test: list = []
        self.X_train: list = []
        self.coef: list = []

        if get_movement_detection_rate:
            self.mov_detection_rates_test: list = []
            self.tprate_test: list = []
            self.fprate_test: list = []
            self.mov_detection_rates_train: list = []
            self.tprate_train: list = []
            self.fprate_train: list = []
        if RUN_BAY_OPT:
            self.best_bay_opt_params: list = []
        if mrmr_select:
            self.mrmr_select: list = []
        if model_save:
            self.model_save: list = []


class Decoder:
    class ClassMissingException(Exception):
        def __init__(
            self,
            message="Only one class present.",
        ) -> None:
            self.message = message
            super().__init__(self.message)

        def __str__(self):
            return self.message

    def __init__(
        self,
        features: "pd.DataFrame| None " = None,
        label: np.ndarray | None = None,
        label_name: str | None = None,
        used_chs: list[str] = [],
        model=LinearRegression(),
        eval_method: Callable = r2_score,
        cv_method=model_selection.KFold(n_splits=3, shuffle=False),
        use_nested_cv: bool = False,
        threshold_score=True,
        mov_detection_threshold: float = 0.5,
        TRAIN_VAL_SPLIT: bool = False,
        RUN_BAY_OPT: bool = False,
        STACK_FEATURES_N_SAMPLES: bool = False,
        time_stack_n_samples: int = 5,
        save_coef: bool = False,
        get_movement_detection_rate: bool = False,
        min_consequent_count: int = 3,
        bay_opt_param_space: list = [],
        VERBOSE: bool = False,
        sfreq: int | None = None,
        undersampling: bool = False,
        oversampling: bool = False,
        mrmr_select: bool = False,
        pca: bool = False,
        cca: bool = False,
        model_save: bool = False,
    ) -> None:
        """Initialize here a feature file for processing
        Read settings.json channels.csv and features.csv
        Read target label

        Parameters
        ----------
        model : machine learning model
            model that utilizes fit and predict functions
        eval_method : sklearn metrics
            evaluation scoring method, will default to r2_score if not passed
        cv_method : sklearm model_selection method
        threshold_score : boolean
            if True set lower threshold at zero (useful for r2),
        mov_detection_threshold : float
            if get_movement_detection_rate is True, find given minimum 'threshold' respective
            consecutive movement blocks, by default 0.5
        TRAIN_VAL_SPLIT (boolean):
            if true split data into additinal validation, and run class weighted CV
        save_coef (boolean):
            if true, save model._coef trained coefficients
        get_movement_detection_rate (boolean):
            save detection rate and tpr / fpr as well
        min_consequent_count (int):
            if get_movement_detection_rate is True, find given 'min_consequent_count' respective
            consecutive movement blocks with minimum size of 'min_consequent_count'
        """

        self.model = model
        self.eval_method = eval_method
        self.cv_method = cv_method
        self.use_nested_cv = use_nested_cv
        self.threshold_score = threshold_score
        self.mov_detection_threshold = mov_detection_threshold
        self.TRAIN_VAL_SPLIT = TRAIN_VAL_SPLIT
        self.RUN_BAY_OPT = RUN_BAY_OPT
        self.save_coef = save_coef
        self.sfreq = sfreq
        self.get_movement_detection_rate = get_movement_detection_rate
        self.min_consequent_count = min_consequent_count
        self.STACK_FEATURES_N_SAMPLES = STACK_FEATURES_N_SAMPLES
        self.time_stack_n_samples = time_stack_n_samples
        self.bay_opt_param_space = bay_opt_param_space
        self.VERBOSE = VERBOSE
        self.undersampling = undersampling
        self.oversampling = oversampling
        self.mrmr_select = mrmr_select
        self.used_chs = used_chs
        self.label = label
        self.label_name = label_name
        self.cca = cca
        self.pca = pca
        self.model_save = model_save

        self.set_data(features)

        self.ch_ind_data = {}
        self.grid_point_ind_data = {}
        self.active_gridpoints = []
        self.feature_names = []
        self.ch_ind_results = {}
        self.gridpoint_ind_results = {}
        self.all_ch_results = {}
        self.columns_names_single_ch = None

        if undersampling:
            from imblearn.under_sampling import RandomUnderSampler

            self.rus = RandomUnderSampler(random_state=0)

        if oversampling:
            from imblearn.over_sampling import RandomOverSampler

            self.ros = RandomOverSampler(random_state=0)

    def set_data(self, features):
        if features is not None:
            self.features = features
            self.feature_names = [
                col
                for col in self.features.columns
                if any(col.startswith(used_ch) for used_ch in self.used_chs)
            ]
            self.data = np.nan_to_num(np.array(self.features[self.feature_names]))

    def set_data_ind_channels(self):
        """specified channel individual data"""
        self.ch_ind_data = {}
        for ch in self.used_chs:
            self.ch_ind_data[ch] = np.nan_to_num(
                np.array(
                    self.features[
                        [col for col in self.features.columns if col.startswith(ch)]
                    ]
                )
            )

    def set_CV_results(self, attr_name, contact_point=None):
        """set CV results in respectie nm_decode attributes
        The reference is first stored in obj_set, and the used lateron

        Parameters
        ----------
        attr_name : string
            is either all_ch_results, ch_ind_results, gridpoint_ind_results
        contact_point : object, optional
            usually an int specifying the grid_point or string, specifying the used channel,
            by default None
        """
        if contact_point is not None:
            getattr(self, attr_name)[contact_point] = {}
            obj_set = getattr(self, attr_name)[contact_point]
        else:
            obj_set = getattr(self, attr_name)

        def set_scores(cv_res: CV_res, set_inner_CV_res: bool = False):
            """
            This function renames the CV_res keys for InnerCV
            """

            def set_score(key_: str, val):
                if set_inner_CV_res:
                    key_ = "InnerCV_" + key_
                obj_set[key_] = val

            set_score("score_train", cv_res.score_train)
            set_score("score_test", cv_res.score_test)
            set_score("y_test", cv_res.y_test)
            set_score("y_train", cv_res.y_train)
            set_score("y_test_pr", cv_res.y_test_pr)
            set_score("y_train_pr", cv_res.y_train_pr)
            set_score("X_train", cv_res.X_train)
            set_score("X_test", cv_res.X_test)

            if self.save_coef:
                set_score("coef", cv_res.coef)
            if self.get_movement_detection_rate:
                set_score("mov_detection_rates_test", cv_res.mov_detection_rates_test)
                set_score(
                    "mov_detection_rates_train",
                    cv_res.mov_detection_rates_train,
                )
                set_score("fprate_test", cv_res.fprate_test)
                set_score("fprate_train", cv_res.fprate_train)
                set_score("tprate_test", cv_res.tprate_test)
                set_score("tprate_train", cv_res.tprate_train)

            if self.RUN_BAY_OPT:
                set_score("best_bay_opt_params", cv_res.best_bay_opt_params)

            if self.mrmr_select:
                set_score("mrmr_select", cv_res.mrmr_select)
            if self.model_save:
                set_score("model_save", cv_res.model_save)
            return obj_set

        obj_set = set_scores(self.cv_res)

        if self.use_nested_cv:
            obj_set = set_scores(self.cv_res_inner, set_inner_CV_res=True)

    def run_CV_caller(self, feature_contacts: str = "ind_channels"):
        """Wrapper that call for all channels / grid points / combined channels the CV function

        Parameters
        ----------
        feature_contacts : str, optional
            "grid_points", "ind_channels" or "all_channels_combined" , by default "ind_channels"
        """
        valid_feature_contacts = [
            "ind_channels",
            "all_channels_combined",
            "grid_points",
        ]
        if feature_contacts not in valid_feature_contacts:
            raise ValueError(f"{feature_contacts} not in {valid_feature_contacts}")

        if feature_contacts == "grid_points":
            for grid_point in self.active_gridpoints:
                self.run_CV(self.grid_point_ind_data[grid_point], self.label)
                self.set_CV_results("gridpoint_ind_results", contact_point=grid_point)
            return self.gridpoint_ind_results

        if feature_contacts == "ind_channels":
            for ch in self.used_chs:
                self.ch_name_tested = ch
                self.run_CV(self.ch_ind_data[ch], self.label)
                self.set_CV_results("ch_ind_results", contact_point=ch)
            return self.ch_ind_results

        if feature_contacts == "all_channels_combined":
            dat_combined = np.array(self.data)
            self.run_CV(dat_combined, self.label)
            self.set_CV_results("all_ch_results", contact_point=None)
            return self.all_ch_results

    def set_data_grid_points(self, cortex_only=False, subcortex_only=False):
        """Read the run_analysis
        Projected data has the shape (samples, grid points, features)
        """

        # activate_gridpoints stores cortex + subcortex data
        self.active_gridpoints = np.unique(
            [
                i.split("_")[0] + "_" + i.split("_")[1]
                for i in self.features.columns
                if "grid" in i
            ]
        )

        if cortex_only:
            self.active_gridpoints = [
                i for i in self.active_gridpoints if i.startswith("gridcortex")
            ]

        if subcortex_only:
            self.active_gridpoints = [
                i for i in self.active_gridpoints if i.startswith("gridsubcortex")
            ]

        self.feature_names = [
            i[len(self.active_gridpoints[0] + "_") :]
            for i in self.features.columns
            if self.active_gridpoints[0] + "_" in i
        ]

        self.grid_point_ind_data = {}

        self.grid_point_ind_data = {
            grid_point: np.nan_to_num(
                self.features[
                    [i for i in self.features.columns if grid_point + "_" in i]
                ]
            )
            for grid_point in self.active_gridpoints
        }

    def get_movement_grouped_array(
        self, prediction, threshold=0.5, min_consequent_count=5
    ):
        """Return given a 1D numpy array, an array of same size with grouped consective blocks

        Parameters
        ----------
        prediction : np.ndarray
            numpy array of either predictions or labels, that is going to be grouped
        threshold : float, optional
            threshold to be applied to 'prediction', by default 0.5
        min_consequent_count : int, optional
            minimum required consective samples higher than 'threshold', by default 5

        Returns
        -------
        labeled_array : np.ndarray
            grouped vector with incrementing number for movement blocks
        labels_count : int
            count of individual movement blocks
        """

        from scipy.ndimage import label as label_ndimage
        from scipy.ndimage import binary_dilation, binary_erosion

        mask = prediction > threshold
        structure = [True] * min_consequent_count  # used for erosion and dilation
        eroded = binary_erosion(mask, structure)
        dilated = binary_dilation(eroded, structure)
        labeled_array, labels_count = label_ndimage(dilated)
        return labeled_array, labels_count

    def calc_movement_detection_rate(
        self, y_label, prediction, threshold=0.5, min_consequent_count=3
    ):
        """Given a label and prediction, return the movement detection rate on the basis of
        movements classified in blocks of 'min_consequent_count'.

        Parameters
        ----------
        y_label : [type]
            [description]
        prediction : [type]
            [description]
        threshold : float, optional
            threshold to be applied to 'prediction', by default 0.5
        min_consequent_count : int, optional
            minimum required consective samples higher than 'threshold', by default 3

        Returns
        -------
        mov_detection_rate : float
            movement detection rate, where at least 'min_consequent_count' samples where high in prediction
        fpr : np.ndarray
            sklearn.metrics false positive rate np.ndarray
        tpr : np.ndarray
            sklearn.metrics true positive rate np.ndarray
        """
        from sklearn.metrics import confusion_matrix

        pred_grouped, _ = self.get_movement_grouped_array(
            prediction, threshold, min_consequent_count
        )
        y_grouped, labels_count = self.get_movement_grouped_array(
            y_label, threshold, min_consequent_count
        )

        hit_rate = np.zeros(labels_count)
        pred_group_bin = np.array(pred_grouped > 0)

        for label_number in range(1, labels_count + 1):  # labeling starts from 1
            hit_rate[label_number - 1] = np.sum(
                pred_group_bin[np.where(y_grouped == label_number)[0]]
            )

        try:
            mov_detection_rate = np.where(hit_rate > 0)[0].shape[0] / labels_count
        except ZeroDivisionError:
            logger.warning("no movements in label")
            return 0, 0, 0

        # calculating TPR and FPR: https://stackoverflow.com/a/40324184/5060208
        CM = confusion_matrix(y_label, prediction)

        TN = CM[0][0]
        FN = CM[1][0]
        TP = CM[1][1]
        FP = CM[0][1]
        fpr = FP / (FP + TN)
        tpr = TP / (TP + FN)

        return mov_detection_rate, fpr, tpr

    def init_cv_res(self) -> None:
        return CV_res(
            get_movement_detection_rate=self.get_movement_detection_rate,
            RUN_BAY_OPT=self.RUN_BAY_OPT,
            mrmr_select=self.mrmr_select,
            model_save=self.model_save,
        )

    # @staticmethod
    # @jit(nopython=True)
    def append_previous_n_samples(X: np.ndarray, y: np.ndarray, n: int = 5):
        """
        stack feature vector for n samples
        """
        TIME_DIM = X.shape[0] - n
        FEATURE_DIM = int(n * X.shape[1])
        time_arr = np.empty((TIME_DIM, FEATURE_DIM))
        for time_idx, time_ in enumerate(np.arange(n, X.shape[0])):
            for time_point in range(n):
                time_arr[
                    time_idx,
                    time_point * X.shape[1] : (time_point + 1) * X.shape[1],
                ] = X[time_ - time_point, :]
        return time_arr, y[n:]

    @staticmethod
    def append_samples_val(X_train, y_train, X_val, y_val, n):
        X_train, y_train = Decoder.append_previous_n_samples(X_train, y_train, n=n)
        X_val, y_val = Decoder.append_previous_n_samples(X_val, y_val, n=n)
        return X_train, y_train, X_val, y_val

    def fit_model(self, model, X_train, y_train):
        if self.TRAIN_VAL_SPLIT:
            X_train, X_val, y_train, y_val = model_selection.train_test_split(
                X_train, y_train, train_size=0.7, shuffle=False
            )

            if y_train.sum() == 0 or y_val.sum(0) == 0:
                raise Decoder.ClassMissingException

            # if type(model) is xgboost.sklearn.XGBClassifier:
            #     classes_weights = class_weight.compute_sample_weight(
            #         class_weight="balanced", y=y_train
            #     )
            #     model.set_params(eval_metric="logloss")
            #     model.fit(
            #         X_train,
            #         y_train,
            #         eval_set=[(X_val, y_val)],
            #         early_stopping_rounds=7,
            #         sample_weight=classes_weights,
            #         verbose=self.VERBOSE,
            #     )
            # elif type(model) is xgboost.sklearn.XGBRegressor:
            #     # might be necessary to adapt for other classifiers
            #
            #     def evalerror(preds, dtrain):
            #         labels = dtrain.get_label()
            #         # return a pair metric_name, result. The metric name must not contain a
            #         # colon (:) or a space since preds are margin(before logistic
            #         # transformation, cutoff at 0)
            #
            #         r2 = metrics.r2_score(labels, preds)
            #
            #         if r2 < 0:
            #             r2 = 0
            #
            #         return "r2", -r2
            #
            #     model.set_params(eval_metric=evalerror)
            #     model.fit(
            #         X_train,
            #         y_train,
            #         eval_set=[(X_val, y_val)],
            #         early_stopping_rounds=10,
            #         verbose=self.VERBOSE,
            #     )
            # else:
            #     model.fit(X_train, y_train, eval_set=[(X_val, y_val)])
        else:
            # check for LDA; and apply rebalancing
            if self.oversampling:
                X_train, y_train = self.ros.fit_resample(X_train, y_train)
            if self.undersampling:
                X_train, y_train = self.rus.fit_resample(X_train, y_train)

            # if type(model) is xgboost.sklearn.XGBClassifier:
            #    model.set_params(eval_metric="logloss")
            #    model.fit(X_train, y_train)
            # else:
            model.fit(X_train, y_train)

        return model

    def eval_model(
        self,
        model_train,
        X_train,
        X_test,
        y_train,
        y_test,
        cv_res: CV_res,
        save_data=True,
        save_probabilities=False,
    ) -> CV_res:
        if self.save_coef:
            cv_res.coef.append(model_train.coef_)

        y_test_pr = model_train.predict(X_test)
        y_train_pr = model_train.predict(X_train)

        sc_te = self.eval_method(y_test, y_test_pr)
        sc_tr = self.eval_method(y_train, y_train_pr)

        if self.threshold_score:
            if sc_tr < 0:
                sc_tr = 0
            if sc_te < 0:
                sc_te = 0

        if self.get_movement_detection_rate:
            self._set_movement_detection_rates(
                y_test, y_test_pr, y_train, y_train_pr, cv_res
            )

        cv_res.score_train.append(sc_tr)
        cv_res.score_test.append(sc_te)
        if save_data:
            cv_res.X_train.append(X_train)
            cv_res.X_test.append(X_test)
        if self.model_save:
            cv_res.model_save.append(deepcopy(model_train))  # clone won't copy params
        cv_res.y_train.append(y_train)
        cv_res.y_test.append(y_test)

        if not save_probabilities:
            cv_res.y_train_pr.append(y_train_pr)
            cv_res.y_test_pr.append(y_test_pr)
        else:
            cv_res.y_train_pr.append(model_train.predict_proba(X_train))
            cv_res.y_test_pr.append(model_train.predict_proba(X_test))
        return cv_res

    def _set_movement_detection_rates(
        self,
        y_test: np.ndarray,
        y_test_pr: np.ndarray,
        y_train: np.ndarray,
        y_train_pr: np.ndarray,
        cv_res: CV_res,
    ) -> CV_res:
        mov_detection_rate, fpr, tpr = self.calc_movement_detection_rate(
            y_test,
            y_test_pr,
            self.mov_detection_threshold,
            self.min_consequent_count,
        )

        cv_res.mov_detection_rates_test.append(mov_detection_rate)
        cv_res.tprate_test.append(tpr)
        cv_res.fprate_test.append(fpr)

        mov_detection_rate, fpr, tpr = self.calc_movement_detection_rate(
            y_train,
            y_train_pr,
            self.mov_detection_threshold,
            self.min_consequent_count,
        )

        cv_res.mov_detection_rates_train.append(mov_detection_rate)
        cv_res.tprate_train.append(tpr)
        cv_res.fprate_train.append(fpr)

        return cv_res

    def wrapper_model_train(
        self,
        X_train,
        y_train,
        X_test=None,
        y_test=None,
        cv_res: CV_res | None = None,
        return_fitted_model_only: bool = False,
        save_data=True,
    ):
        if cv_res is None:
            cv_res = CV_res(
                get_movement_detection_rate=self.get_movement_detection_rate,
                RUN_BAY_OPT=self.RUN_BAY_OPT,
                mrmr_select=self.mrmr_select,
                model_save=self.model_save,
            )

        model_train = clone(self.model)
        if self.STACK_FEATURES_N_SAMPLES:
            if X_test is not None:
                X_train, y_train, X_test, y_test = Decoder.append_samples_val(
                    X_train,
                    y_train,
                    X_test,
                    y_test,
                    n=self.time_stack_n_samples,
                )
            else:
                X_train, y_train = Decoder.append_previous_n_samples(
                    X_train, y_train, n=self.time_stack_n_samples
                )

        if y_train.sum() == 0 or (
            y_test is not None and y_test.sum() == 0
        ):  # only one class present
            raise Decoder.ClassMissingException

        if self.RUN_BAY_OPT:
            model_train = self.bay_opt_wrapper(model_train, X_train, y_train)

        if self.mrmr_select:
            from mrmr import mrmr_classif

            if len(self.feature_names) > X_train.shape[1]:
                # analyze induvidual ch
                columns_names = [
                    col
                    for col in self.feature_names
                    if col.startswith(self.ch_name_tested)
                ]
                if self.columns_names_single_ch is None:
                    self.columns_names_single_ch = [
                        f[len(self.ch_name_tested) + 1 :] for f in columns_names
                    ]
            else:
                # analyze all_ch_combined
                columns_names = self.feature_names
            X_train = pd.DataFrame(X_train, columns=columns_names)
            X_test = pd.DataFrame(X_test, columns=columns_names)

            y_train = pd.Series(y_train)
            selected_features = mrmr_classif(X=X_train, y=y_train, K=20, n_jobs=60)

            X_train = X_train[selected_features]
            X_test = X_test[selected_features]

        if self.pca:
            from sklearn.decomposition import PCA

            pca = PCA(n_components=10)
            pca.fit(X_train)
            X_train = pca.transform(X_train)
            X_test = pca.transform(X_test)

        if self.cca:
            from sklearn.cross_decomposition import CCA

            cca = CCA(n_components=10)
            cca.fit(X_train, y_train)
            X_train = cca.transform(X_train)
            X_test = cca.transform(X_test)

        if self.STACK_FEATURES_N_SAMPLES:
            if return_fitted_model_only:
                X_train, y_train = self.append_previous_n_samples(
                    X_train, y_train, self.time_stack_n_samples
                )
            else:
                X_train, y_train, X_test, y_test = self.append_samples_val(
                    X_train, y_train, X_test, y_test, self.time_stack_n_samples
                )

        # fit model
        model_train = self.fit_model(model_train, X_train, y_train)

        if return_fitted_model_only:
            return model_train

        cv_res = self.eval_model(
            model_train, X_train, X_test, y_train, y_test, cv_res, save_data
        )

        if self.mrmr_select:
            cv_res.mrmr_select.append(selected_features)

        return cv_res

    def run_CV(self, data, label):
        """Evaluate model performance on the specified cross validation.
        If no data and label is specified, use whole feature class attributes.

        Parameters
        ----------
        data (np.ndarray):
            data to train and test with shape samples, features
        label (np.ndarray):
            label to train and test with shape samples, features
        """

        def split_data(data):
            if self.cv_method == "NonShuffledTrainTestSplit":
                # set outer 10s set to train index
                # test index is thus in the middle starting at random number
                N_samples = data.shape[0]
                test_area_points = (N_samples - self.sfreq * 10) - (self.sfreq * 10)
                test_points = int(N_samples * 0.3)

                if test_area_points > test_points:
                    start_index = np.random.randint(
                        int(self.sfreq * 10),
                        N_samples - self.sfreq * 10 - test_points,
                    )
                    test_index = np.arange(start_index, start_index + test_points)
                    train_index = np.concatenate(
                        (
                            np.arange(0, start_index),
                            np.arange(start_index + test_points, N_samples),
                        ),
                        axis=0,
                    ).flatten()
                    yield train_index, test_index
                else:
                    cv_single_tr_te_split = model_selection.check_cv(
                        cv=[
                            model_selection.train_test_split(
                                np.arange(data.shape[0]),
                                test_size=0.3,
                                shuffle=False,
                            )
                        ]
                    )
                    for (
                        train_index,
                        test_index,
                    ) in cv_single_tr_te_split.split():
                        yield train_index, test_index
            else:
                for train_index, test_index in self.cv_method.split(data):
                    yield train_index, test_index

        cv_res = self.init_cv_res()

        if self.use_nested_cv:
            cv_res_inner = self.init_cv_res()

        for train_index, test_index in split_data(data):
            X_train, y_train = data[train_index, :], label[train_index]
            X_test, y_test = data[test_index], label[test_index]
            try:
                cv_res = self.wrapper_model_train(
                    X_train, y_train, X_test, y_test, cv_res
                )
            except Decoder.ClassMissingException:
                continue

            if self.use_nested_cv:
                data_inner = data[train_index]
                label_inner = label[train_index]
                for train_index_inner, test_index_inner in split_data(data_inner):
                    X_train_inner = data_inner[train_index_inner, :]
                    y_train_inner = label_inner[train_index_inner]
                    X_test_inner = data_inner[test_index_inner]
                    y_test_inner = label_inner[test_index_inner]
                    try:
                        cv_res_inner = self.wrapper_model_train(
                            X_train_inner,
                            y_train_inner,
                            X_test_inner,
                            y_test_inner,
                            cv_res_inner,
                        )
                    except Decoder.ClassMissingException:
                        continue

        self.cv_res = cv_res
        if self.use_nested_cv:
            self.cv_res_inner = cv_res_inner

    def bay_opt_wrapper(self, model_train, X_train, y_train):
        """Run bayesian optimization and test best params to model_train
        Save best params into self.best_bay_opt_params
        """

        (
            X_train_bo,
            X_test_bo,
            y_train_bo,
            y_test_bo,
        ) = model_selection.train_test_split(
            X_train, y_train, train_size=0.7, shuffle=False
        )

        if y_train_bo.sum() == 0 or y_test_bo.sum() == 0:
            logger.critical("could not start Bay. Opt. with no labels > 0")
            raise Decoder.ClassMissingException

        params_bo = self.run_Bay_Opt(
            X_train_bo, y_train_bo, X_test_bo, y_test_bo, rounds=10
        )

        # set bay. opt. obtained best params to model
        params_bo_dict = {}
        for i in range(len(params_bo)):
            setattr(model_train, self.bay_opt_param_space[i].name, params_bo[i])
            params_bo_dict[self.bay_opt_param_space[i].name] = params_bo[i]

        self.best_bay_opt_params.append(params_bo_dict)

        return model_train

    def run_Bay_Opt(
        self,
        X_train,
        y_train,
        X_test,
        y_test,
        rounds=30,
        base_estimator="GP",
        acq_func="EI",
        acq_optimizer="sampling",
        initial_point_generator="lhs",
    ):
        """Run skopt bayesian optimization
        skopt.Optimizer:
        https://scikit-optimize.github.io/stable/modules/generated/skopt.Optimizer.html#skopt.Optimizer

        example:
        https://scikit-optimize.github.io/stable/auto_examples/ask-and-tell.html#sphx-glr-auto-examples-ask-and-tell-py

        Special attention needs to be made with the run_CV output,
        some metrics are minimized (MAE), some are maximized (r^2)

        Parameters
        ----------
        X_train: np.ndarray
        y_train: np.ndarray
        X_test: np.ndarray
        y_test: np.ndarray
        rounds : int, optional
            optimizing rounds, by default 10
        base_estimator : str, optional
            surrogate model, used as optimization function instead of cross validation, by default "GP"
        acq_func : str, optional
            function to minimize over the posterior distribution, by default "EI"
        acq_optimizer : str, optional
            method to minimize the acquisition function, by default "sampling"
        initial_point_generator : str, optional
            sets a initial point generator, by default "lhs"

        Returns
        -------
        skopt result parameters
        """

        def get_f_val(model_bo):
            try:
                model_bo = self.fit_model(model_bo, X_train, y_train)
            except Decoder.ClassMissingException:
                pass

            return self.eval_method(y_test, model_bo.predict(X_test))

        from skopt import Optimizer

        opt = Optimizer(
            self.bay_opt_param_space,
            base_estimator=base_estimator,
            acq_func=acq_func,
            acq_optimizer=acq_optimizer,
            initial_point_generator=initial_point_generator,
        )

        for _ in range(rounds):
            next_x = opt.ask()
            # set model values
            model_bo = clone(self.model)
            for i in range(len(next_x)):
                setattr(model_bo, self.bay_opt_param_space[i].name, next_x[i])
            f_val = get_f_val(model_bo)
            res = opt.tell(next_x, f_val)
            if self.VERBOSE:
                logger.info(f_val)

        # res is here automatically appended by skopt
        return res.x

    def save(self, feature_path: str, feature_file: str, str_save_add=None) -> None:
        """Save decoder object to pickle"""

        # why is the decoder not saved to a .json?

        if str_save_add is None:
            PATH_OUT = PurePath(feature_path, feature_file, feature_file + "_ML_RES.p")
        else:
            PATH_OUT = PurePath(
                feature_path,
                feature_file,
                feature_file + "_" + str_save_add + "_ML_RES.p",
            )

        logger.info(f"model being saved to: {PATH_OUT}")
        with open(PATH_OUT, "wb") as output:  # Overwrites any existing file.
            pickle.dump(self, output)



from collections.abc import Iterable
from pydantic import field_validator
from typing import TYPE_CHECKING, Callable

import numpy as np

from py_neuromodulation.utils.types import (
    NMBaseModel,
    NMFeature,
    BoolSelector,
    FrequencyRange,
)

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


class BispectraComponents(BoolSelector):
    absolute: bool = True
    real: bool = True
    imag: bool = True
    phase: bool = True


class BispectraFeatures(BoolSelector):
    mean: bool = True
    sum: bool = True
    var: bool = True


class BispectraSettings(NMBaseModel):
    f1s: FrequencyRange = FrequencyRange(5, 35)
    f2s: FrequencyRange = FrequencyRange(5, 35)
    compute_features_for_whole_fband_range: bool = True
    frequency_bands: list[str] = ["theta", "alpha", "low_beta", "high_beta"]

    components: BispectraComponents = BispectraComponents()
    bispectrum_features: BispectraFeatures = BispectraFeatures()

    @field_validator("f1s", "f2s")
    def test_range(cls, filter_range):
        assert (
            filter_range[1] > filter_range[0]
        ), f"second frequency range value needs to be higher than first one, got {filter_range}"
        return filter_range

    @field_validator("frequency_bands")
    def fbands_spaces_to_underscores(cls, frequency_bands):
        return [f.replace(" ", "_") for f in frequency_bands]


FEATURE_DICT: dict[str, Callable] = {
    "mean": np.nanmean,
    "sum": np.nansum,
    "var": np.nanvar,
}

COMPONENT_DICT: dict[str, Callable] = {
    "real": lambda obj: getattr(obj, "real"),
    "imag": lambda obj: getattr(obj, "imag"),
    "absolute": np.abs,
    "phase": np.angle,
}


class Bispectra(NMFeature):
    def __init__(
        self, settings: "NMSettings", ch_names: Iterable[str], sfreq: float
    ) -> None:
        self.sfreq = sfreq
        self.ch_names = ch_names
        self.frequency_ranges_hz = settings.frequency_ranges_hz
        self.settings: BispectraSettings = settings.bispectrum_settings

        assert all(
            f_band_bispectrum in settings.frequency_ranges_hz
            for f_band_bispectrum in self.settings.frequency_bands
        ), (
            "bispectrum selected frequency bands don't match the ones"
            "specified in s['frequency_ranges_hz']"
            f"bispectrum frequency bands: {self.settings.frequency_bands}"
            f"specified frequency_ranges_hz: {settings.frequency_ranges_hz}"
        )

        self.used_features = self.settings.bispectrum_features.get_enabled()

        self.min_freq = min(
            self.settings.f1s.frequency_low_hz, self.settings.f2s.frequency_low_hz
        )
        self.max_freq = max(
            self.settings.f1s.frequency_high_hz, self.settings.f2s.frequency_high_hz
        )

        # self.freqs: np.ndarray = np.array([]) # In case we pre-computed this

    def calc_feature(self, data: np.ndarray) -> dict:
        from pybispectra import compute_fft, WaveShape

        fft_coeffs, freqs = compute_fft(
            data=np.expand_dims(data, axis=0),
            sampling_freq=self.sfreq,
            n_points=data.shape[1],
            verbose=False,
        )

        # freqs is batch independent, except for the last batch perhaps (if it has different shape)
        # but it's computed by compute_fft regardless so no advantage in pre-computing it
        # if not self.freqs = self.freqs = np.fft.rfftfreq(n=data.shape[1], d = 1 / sfreq)

        # fft_coeffs shape: [epochs, channels, frequencies]

        f_spectrum_range = freqs[
            np.logical_and(freqs >= self.min_freq, freqs <= self.max_freq)
        ]

        waveshape = WaveShape(
            data=fft_coeffs,
            freqs=freqs,
            sampling_freq=self.sfreq,
            verbose=False,
        )

        waveshape.compute(
            f1s=tuple(self.settings.f1s),  # type: ignore
            f2s=tuple(self.settings.f2s),  # type: ignore
        )
        waveshape = waveshape.results.get_results(copy=False)  # can overwrite obj with array

        feature_results = {}
        for ch_idx, ch_name in enumerate(self.ch_names):
            bispectrum = waveshape[ch_idx]

            for component in self.settings.components.get_enabled():
                spectrum_ch = COMPONENT_DICT[component](bispectrum)

                for fb in self.settings.frequency_bands:
                    range_ = (f_spectrum_range >= self.frequency_ranges_hz[fb][0]) & (
                        f_spectrum_range <= self.frequency_ranges_hz[fb][1]
                    )
                    # waveshape.results.plot()
                    data_bs = spectrum_ch[range_, range_]

                    for bispectrum_feature in self.used_features:
                        feature_results[
                            f"{ch_name}_Bispectrum_{component}_{bispectrum_feature}_{fb}"
                        ] = FEATURE_DICT[bispectrum_feature](data_bs)

                        if self.settings.compute_features_for_whole_fband_range:
                            feature_results[
                                f"{ch_name}_Bispectrum_{component}_{bispectrum_feature}_whole_fband_range"
                            ] = FEATURE_DICT[bispectrum_feature](spectrum_ch)

        return feature_results



import os
import numpy as np
from threading import Thread
import time
import asyncio
import multiprocessing as mp
from queue import Empty
from pathlib import Path
from py_neuromodulation.stream import Stream, NMSettings
from py_neuromodulation.analysis.decode import RealTimeDecoder
from py_neuromodulation.utils import set_channels
from py_neuromodulation.utils.io import read_mne_data
from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation import logger
from py_neuromodulation.gui.backend.app_socket import WebsocketManager
from py_neuromodulation.stream.backend_interface import StreamBackendInterface
from py_neuromodulation import logger


class PyNMState:
    def __init__(
        self,
        log_queue_size: bool = False,
    ) -> None:
        self.log_queue_size = log_queue_size
        self.lsl_stream_name: str = ""
        self.is_stream_lsl: bool = False
        self.experiment_name: str = "PyNM_Experiment"  # set by set_stream_params
        self.out_dir: _PathLike = str(
            Path.home() / "PyNM" / self.experiment_name
        )  # set by set_stream_params
        self.decoding_model_path: _PathLike | None = None
        self.decoder: RealTimeDecoder | None = None

        self.backend_interface: StreamBackendInterface | None = None
        self.websocket_manager: WebsocketManager | None = None

        # Note: sfreq and data are required for stream init
        self.stream: Stream = Stream(sfreq=1500, data=np.random.random([1, 1]))

        self.feature_queue = mp.Queue()
        self.rawdata_queue = mp.Queue()
        self.control_queue = mp.Queue()
        self.stop_event = asyncio.Event()

        self.messages_sent = 0

    def start_run_function(
        self,
        websocket_manager: WebsocketManager | None = None,
    ) -> None:
        # Clear stop event to enable _process_queue
        self.stop_event.clear()
        self.websocket_manager = websocket_manager

        # Create decoder
        if self.decoding_model_path is not None and self.decoding_model_path != "None":
            if os.path.exists(self.decoding_model_path):
                self.decoder = RealTimeDecoder(self.decoding_model_path)
            else:
                logger.debug("Passed decoding model path does't exist")

        # Initialize the backend interface if not already done
        if not self.backend_interface:
            self.backend_interface = StreamBackendInterface(
                self.feature_queue, self.rawdata_queue, self.control_queue
            )

        # The run_func_thread is terminated through the stream_handling_queue
        # which initiates to break the data generator and save the features
        stream_process = mp.Process(
            target=self.stream.run,
            kwargs={
                "out_dir": "" if self.out_dir == "default" else self.out_dir,
                "experiment_name": self.experiment_name,
                "is_stream_lsl": self.is_stream_lsl,
                "stream_lsl_name": self.lsl_stream_name,
                "simulate_real_time": True,
                "decoder": self.decoder,
                "backend_interface": self.backend_interface,
            },
        )

        stream_process.start()

        # Start websocket sender process

        if self.websocket_manager:
            # TONI: Instead of having this function be not async and send the
            # _process_queue function to the Uvicorn async loop, we could
            # have this entire "start_run_function" function be async as well

            # Get the current event loop and run the queue processor
            loop = asyncio.get_running_loop()
            queue_task = loop.create_task(self._process_queue())

            # Store task reference for cleanup
            self._queue_task = queue_task

        # Store processes for cleanup
        self.stream_process = stream_process

    def stop_run_function(self) -> None:
        """Stop the stream processing"""
        if self.backend_interface:
            self.backend_interface.send_command("stop")
            self.stop_event.set()

    def setup_lsl_stream(
        self,
        lsl_stream_name: str = "",
        line_noise: float | None = None,
    ):
        from mne_lsl.lsl import resolve_streams

        logger.info("resolving streams")
        logger.info(f"passed lsl_stream_name: {lsl_stream_name}")
        lsl_streams = resolve_streams()

        for stream in lsl_streams:
            if stream.name == lsl_stream_name:
                logger.info(f"found stream {lsl_stream_name}")

                ch_names = stream.get_channel_names()
                if ch_names is None:
                    ch_names = ["ch" + str(i) for i in range(stream.n_channels)]
                logger.info(f"channel names: {ch_names}")

                ch_types = stream.get_channel_types()
                if ch_types is None:
                    ch_types = ["eeg" for i in range(stream.n_channels)]

                logger.info(f"channel types: {ch_types}")

                info_ = stream.get_channel_info()
                logger.info(f"channel info: {info_}")

                channels = set_channels(
                    ch_names=ch_names,
                    ch_types=ch_types,
                    used_types=["eeg", "ecog", "dbs", "seeg"],
                )

                # set all used column to 0
                #channels.loc[:, "used"] = 0

                logger.info(f"channels: {channels}")
                sfreq = stream.sfreq
                logger.info(f"sfreq: {sfreq}")

                logger.info("Starting setup LSL stream...")
                try:
                    self.stream: Stream = Stream(
                        sfreq=sfreq,
                        line_noise=line_noise,
                        channels=channels,
                    )
                except Exception as e:
                    logger.error(f"Error setting up LSL stream: {e}")

                logger.info("Finish setup LSL stream...")
                logger.info("stream setup")
                logger.info("settings setup")

                self.lsl_stream_name = lsl_stream_name
                self.is_stream_lsl = True
                break
        else:
            logger.error(f"Stream {lsl_stream_name} not found")
            self.is_stream_lsl = False
            self.is_stream_lsl = ""
            raise ValueError(f"Stream {lsl_stream_name} not found")

    def setup_offline_stream(
        self,
        file_path: str,
        line_noise: float,
    ):
        data, sfreq, ch_names, ch_types, bads = read_mne_data(file_path)

        channels = set_channels(
            ch_names=ch_names,
            ch_types=ch_types,
            bads=bads,
            reference=None,
            used_types=["eeg", "ecog", "dbs", "seeg"],
            target_keywords=None,
        )

        self.stream: Stream = Stream(
            sfreq=sfreq,
            data=data,
            channels=channels,
            line_noise=line_noise,
        )
        self.is_stream_lsl = False
        self.lsl_stream_name = ""

    # Async function that will continuously run in the Uvicorn async loop
    # and handle sending data through the websocket manager
    async def _process_queue(self):
        last_queue_check = time.time()

        while not self.stop_event.is_set():
            # Use asyncio.gather to process both queues concurrently
            tasks = []
            current_time = time.time()

            # Check feature queue
            while not self.feature_queue.empty():
                try:
                    data = self.feature_queue.get_nowait()
                    tasks.append(self.websocket_manager.send_cbor(data))  # type: ignore
                    self.messages_sent += 1
                except Empty:
                    break

            # Check raw data queue
            while not self.rawdata_queue.empty():
                try:
                    data = self.rawdata_queue.get_nowait()
                    self.messages_sent += 1
                    tasks.append(self.websocket_manager.send_cbor(data))  # type: ignore
                except Empty:
                    break

            if tasks:
                # Wait for all send operations to complete
                await asyncio.gather(*tasks, return_exceptions=True)
            else:
                # Only sleep if we didn't process any messages
                await asyncio.sleep(0.001)

            # Log queue diagnostics every 5 seconds
            if self.log_queue_size:
                if current_time - last_queue_check > 5:
                    logger.info(
                        "\nQueue diagnostics:\n"
                        f"\tMessages send to websocket: {self.messages_sent}.\n"
                    )
                    try:
                        logger.info(
                            f"\tFeature queue size: ~{self.feature_queue.qsize()}\n"
                            f"\tRaw data queue size: ~{self.rawdata_queue.qsize()}"
                        )
                    except NotImplementedError:
                        continue

                last_queue_check = current_time

            # Check if stream process is still alive
            if not self.stream_process.is_alive():
                break



from typing import Literal

import numpy as np
import pytest
from mne_connectivity import make_signals_in_freq_bands

import py_neuromodulation as nm


@pytest.mark.parametrize("method", ["coh", "dpli"])
def test_mne_connectivity(method: Literal["coh", "dpli"]):
    """Check that mne_connectivity features compute properly and match expected values."""
    # Simulate connectivity data (interaction at specified frequency band)
    sfreq = 500  # Hz
    n_epochs = 1
    n_times = sfreq * 2  # samples
    fband = (15, 20)  # frequency band of interaction, Hz
    trans = 2  # transition bandwidth of signal, Hz
    delay = 5  # samples; seed leads target to test directed connectivity methods
    epochs = make_signals_in_freq_bands(
        n_seeds=1,
        n_targets=1,
        freq_band=fband,
        n_epochs=n_epochs,
        n_times=n_times,
        sfreq=sfreq,
        trans_bandwidth=trans,
        connection_delay=delay,
        ch_names=["seed", "target"],
        snr=0.7,  # change here requires change in `signal/noise_con` vars below
        rng_seed=44
    )

    # Set up py_nm channels info
    ch_names = epochs.ch_names
    ch_types = epochs.get_channel_types()
    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=None,
        new_names="default",
        used_types=tuple(np.unique(ch_types)),
        target_keywords=None,
    )

    # Set up pn_nm processing settings
    settings = nm.NMSettings.get_default()
    settings.reset()
    settings.preprocessing = []  # No preprocessing (changes properties of simulated data)
    settings.features.mne_connectivity = True
    settings.segment_length_features_ms = n_times / sfreq * 1000  # use all data at once

    # redefine freq. bands of interest
    # (accounts for signal-noise transition bandwdith when defining frequencies)
    settings.frequency_ranges_hz = {
        "signal": {  # strong connectivity expected
            "frequency_low_hz": fband[0],
            "frequency_high_hz": fband[1],
        },
        "noise_low": {  # weak connectivity expected from 1 Hz to start of interaction
            "frequency_low_hz": 1,
            "frequency_high_hz": fband[0] - trans * 2,
        },
        "noise_high": {  # weak connectivity expected from end of interaction to Nyquist
            "frequency_low_hz": fband[1] + trans * 2,
            "frequency_high_hz": sfreq // 2 - 1,
        },
    }

    settings.mne_connectivity_settings.method = method

    # unique all-to-all connectivity indices
    if method == "coh":  # undirected, so unique combination is [[0, 1]]
        settings.mne_connectivity_settings.channels = [ch_names]
    else:  # directed, so unique combinations are [[0, 1], [1, 0]]
        settings.mne_connectivity_settings.channels = [ch_names, ch_names[::-1]]

    # do not normalise features for this test!
    # (normalisation changes interpretability of connectivity values, making it harder to
    # define 'expected' connectivity values)
    settings.postprocessing.feature_normalization = False

    # Set up py_nm stream
    stream = nm.Stream(
        settings=settings,
        channels=channels,
        path_grids=None,
        verbose=True,
        sfreq=epochs.info["sfreq"],
    )

    # Compute connectivity
    features = stream.run(
        epochs.get_data(copy=False)[0],  # extract first (and only) epoch from obj
        out_dir="./test_data",
        experiment_name="test_mne_connectivity",
    )

    # Aggregate results over windows
    results = {key: None for key in features.keys()}
    results.pop("time")
    for key in results.keys():
        # average over windows; take absolute before averaging icoh values
        results[key] = np.abs(features[key].values).mean()

    if method == "coh":
        # Define expected connectivity values for signal and noise frequencies
        noise_con = 0.4
        signal_con = 0.8

        # Assert that frequencies of simulated interaction have strong connectivity
        np.testing.assert_array_less(
            signal_con, results["coh_seed_to_target_mean_fband_signal"]
        )
        # Assert that frequencies of noise have weak connectivity
        np.testing.assert_array_less(
            results["coh_seed_to_target_mean_fband_noise_low"], noise_con
        )
        np.testing.assert_array_less(
            results["coh_seed_to_target_mean_fband_noise_high"], noise_con
        )
    else:
        # Define expected connectivity values for signal and noise frequencies
        seed_leading = 0.65
        target_leading = 0.35

        # Assert that frequencies of simulated interaction have strong connectivity, i.e.:
        # >> 0.5 for seed-to-target; << 0.5 for target-to-seed
        np.testing.assert_array_less(
            seed_leading, results["dpli_seed_to_target_mean_fband_signal"]
        )
        np.testing.assert_array_less(
            results["dpli_target_to_seed_mean_fband_signal"], target_leading
        )
        # Assert that frequencies of noise have weak connectivity (i.e., ~0.5)
        for con_name in ["seed_to_target", "target_to_seed"]:
            np.testing.assert_array_less(
                results[f"dpli_{con_name}_mean_fband_noise_low"], seed_leading
            )
            np.testing.assert_array_less(
                target_leading, results[f"dpli_{con_name}_mean_fband_noise_low"]
            )



import py_neuromodulation as nm
import numpy as np


def get_fast_compute_settings():
    settings = nm.NMSettings.get_fast_compute()
    settings.preprocessing = ["re_referencing", "notch_filter"]
    settings.features.fft = True
    settings.postprocessing.feature_normalization = True
    return settings


def get_features(time_end_ms: int, segment_length_features_ms: int):
    data = np.random.random([2, time_end_ms])
    settings = get_fast_compute_settings()
    settings.segment_length_features_ms = segment_length_features_ms
    settings.fft_settings.windowlength_ms = segment_length_features_ms

    settings.frequency_ranges_hz = {
        # "high beta" : [20, 35],
        "low_gamma": [60, 80],
        "high_gamma": [90, 200],
    }

    stream = nm.Stream(
        sfreq=1000,
        data=data,
        sampling_rate_features_hz=10,
        verbose=False,
        settings=settings,
    )

    features = stream.run(
        data, out_dir="./test_data", experiment_name="test_feature_sampling_rates"
    )
    return features


def test_feature_timing_5s_start_800ms():
    """Test if the features timing duration are computed correctly for a 5s signal with 800ms segment length"""
    time_end_ms = 5000
    segment_length_features_ms = 800
    features = get_features(time_end_ms, segment_length_features_ms)

    # TONI: this is failing now because I changed the way timestamps are computed
    assert int(features.time.iloc[0]) == segment_length_features_ms

    assert int(features.time.iloc[-1]) == time_end_ms


def test_feature_timing_1s_start_500ms():
    """Test if the features timing duration are computed correctly for a 1s signal with 500ms segment length"""
    time_end_ms = 1000
    segment_length_features_ms = 500
    features = get_features(time_end_ms, segment_length_features_ms)

    # TONI: this is failing now because I changed the way timestamps are computed
    assert int(features.time.iloc[0]) == segment_length_features_ms

    assert int(features.time.iloc[-1]) == time_end_ms



"""Re-referencing Module."""

import numpy as np
import pandas as pd

from py_neuromodulation.utils.types import NMPreprocessor


class ReReferencer(NMPreprocessor):
    def __init__(
        self,
        sfreq: float,
        channels: pd.DataFrame,
    ) -> None:
        """Initialize real-time rereference information.

        Parameters
        ----------
        sfreq : float
            Sampling frequency. Is not used, only kept for compatibility.
        channels : Pandas DataFrame
            Dataframe containing information about rereferencing, as
            specified in channels.csv.


        Raises:
            ValueError: rereferencing using undefined channel
            ValueError: rereferencing to same channel
        """

        self.ref_matrix: np.ndarray | None

        channels = channels[channels["used"] == 1].reset_index(drop=True)
        # (channels_used,) = np.where((channels.used == 1))

        ch_names = channels["name"].tolist()

        # no re-referencing is being performed when there is a single channel present only
        if channels.shape[0] in (0, 1):
            self.ref_matrix = None
            return

        ch_types = channels["type"]
        refs = channels["rereference"]

        type_map = {}
        for ch_type in ch_types.unique():
            type_map[ch_type] = np.where(
                (ch_types == ch_type) & (channels["status"] == "good")
            )[0]

        ref_matrix = np.zeros((len(channels), len(channels)))
        for ind in range(len(channels)):
            ref_matrix[ind, ind] = 1
            # if ind not in channels_used:
            #    continue
            ref = refs[ind]
            if ref.lower() == "none" or pd.isnull(ref):
                ref_idx = None
                continue
            if ref.lower() == "average":
                ch_type = ch_types[ind]
                ref_idx = type_map[ch_type][type_map[ch_type] != ind]
            else:
                ref_idx = []
                ref_channels = ref.split("&")
                for ref_chan in ref_channels:
                    if ref_chan not in ch_names:
                        raise ValueError(
                            "One or more of the reference channels are not"
                            " part of the recording channels. First missing"
                            f" channel: {ref_chan}."
                        )
                    if ref_chan == ch_names[ind]:
                        raise ValueError(
                            "You cannot rereference to the same channel."
                            f" Channel: {ref_chan}."
                        )
                    ref_idx.append(ch_names.index(ref_chan))
            ref_matrix[ind, ref_idx] = -1 / len(ref_idx)
        self.ref_matrix = ref_matrix

    def process(self, data: np.ndarray) -> np.ndarray:
        """Rereference data according to the initialized ReReferencer class.

        Args:
            data (numpy ndarray) :
                shape(n_channels, n_samples) - data to be rereferenced.

        Returns:
            reref_data (numpy ndarray):
                shape(n_channels, n_samples) - rereferenced data
        """
        if self.ref_matrix is not None:
            return self.ref_matrix @ data
        else:
            return data



import py_neuromodulation as nm
import numpy as np


def test_fooof_features(setup_default_stream_fast_compute):
    default_stream: tuple[np.ndarray, nm.Stream] = setup_default_stream_fast_compute
    data, stream = default_stream

    generator = nm.stream.RawDataGenerator(
        data,
        stream.sfreq,
        stream.settings.sampling_rate_features_hz,
        stream.settings.segment_length_features_ms,
    )

    _, data_batch = next(generator, None)
    feature_series = stream.data_processor.process(data_batch)
    # since the settings can define searching for "max_n_peaks" peaks
    # there can be None's in the feature_series
    # with a non successful fit, aperiodic features can also be None
    assert feature_series is not None



import numpy as np

import py_neuromodulation as nm


def test_stream_with_none_data():
    """Test if passing None as the data to a Stream object results in None features."""

    fs = 1000
    data = np.random.random([2, 2000])
    data[0, :] = None

    stream = nm.Stream(sfreq=fs, data=data)

    features = stream.run(
        data, out_dir="./test_data", experiment_name="test_stream_with_none_data"
    )

    # assert if all features if name ch0 are None
    assert len(
        [f for f in features.columns if "ch0" in f and features[f].isna().all()]
    ) == len([f for f in features if "ch0" in f])

    # and check if all features of the second channel are not None
    assert len(
        [f for f in features.columns if "ch1" in f and features[f].notna().all()]
    ) == len([f for f in features if "ch1" in f])



import json
from pathlib import PurePath, Path
from typing import TYPE_CHECKING, cast

import numpy as np

from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation import logger, PYNM_DIR

if TYPE_CHECKING:
    from mne_bids import BIDSPath
    from mne import io as mne_io
    import pandas as pd


def load_channels(
    channels: "pd.DataFrame | _PathLike",
) -> "pd.DataFrame":
    """Read channels from path or specify via BIDS arguments.
    Necessary parameters are then ch_names (list), ch_types (list), bads (list), used_types (list),
    target_keywords (list) and reference Union[list, str].
    """
    import pandas as pd

    if isinstance(channels, pd.DataFrame):
        return channels

    if not Path(channels).is_file():
        raise ValueError("PATH_CHANNELS is not a valid file. Got: " f"{channels}")

    return pd.read_csv(channels)


def read_BIDS_data(
    PATH_RUN: "_PathLike | BIDSPath",
    line_noise: int = 50,
) -> tuple["mne_io.Raw", np.ndarray, float, int, list | None, list | None]:
    """Given a run path and bids data path, read the respective data

    Parameters
    ----------
    PATH_RUN : path to bids run file
        supported formats: https://bids-specification.readthedocs.io/en/v1.2.1/04-modality-specific-files/04-intracranial-electroencephalography.html#ieeg-recording-data
    line_noise: int, optional
        by default 50

    Returns
    -------
    raw_arr : mne.io.RawArray
    raw_arr_data : np.ndarray
    sfreq : float
    line_noise : int
    coord_list : list | None
    coord_names : list | None
    """

    from mne_bids import read_raw_bids, get_bids_path_from_fname

    bids_path = get_bids_path_from_fname(PATH_RUN)

    raw_arr = read_raw_bids(bids_path)
    coord_list, coord_names = get_coord_list(raw_arr)
    if raw_arr.info["line_freq"] is not None:
        line_noise = int(raw_arr.info["line_freq"])
    else:
        logger.info(
            f"Line noise is not available in the data, using value of {line_noise} Hz."
        )
    return (
        raw_arr,
        raw_arr.get_data(),
        raw_arr.info["sfreq"],
        line_noise,
        coord_list,
        coord_names,
    )


def read_mne_data(
    PATH_RUN: "_PathLike | BIDSPath",
    line_noise: int = 50,
) -> tuple[np.ndarray, float, list[str], list[str], list[str]]:
    """Read data in the mne.io.read_raw supported format.

    Parameters
    ----------
    PATH_RUN : _PathLike | BIDSPath
        Path to mne.io.read_raw supported types https://mne.tools/stable/generated/mne.io.read_raw.html
    line_noise : int, optional
        line noise, by default 50

    Returns
    -------
    raw : mne.io.Raw
    sfreq : float
    ch_names : list[str]
    ch_type : list[str]
    bads : list[str]
    """

    from mne import io as mne_io

    raw_arr = mne_io.read_raw(PATH_RUN)
    sfreq = raw_arr.info["sfreq"]
    ch_names = raw_arr.info["ch_names"]
    ch_types = raw_arr.get_channel_types()
    logger.info(
        "Channel data is read using mne.io.read_raw function. Channel types might not be correct"
        " and set to 'eeg' by default"
    )
    bads = raw_arr.info["bads"]

    if raw_arr.info["line_freq"] is not None:
        line_noise = int(raw_arr.info["line_freq"])
    else:
        logger.info(
            f"Line noise is not available in the data, using value of {line_noise} Hz."
        )

    data = cast(np.ndarray, raw_arr.get_data())
    return data, sfreq, ch_names, ch_types, bads


def get_coord_list(
    raw: "mne_io.BaseRaw",
) -> tuple[list, list] | tuple[None, None]:
    """Return the coordinate list and names from mne RawArray

    Parameters
    ----------
    raw : mne_io.BaseRaw

    Returns
    -------
    coord_list[list, list] | coord_names[None, None]
    """
    montage = raw.get_montage()
    if montage is not None:
        coord_list = np.array(
            list(dict(montage.get_positions()["ch_pos"]).values())
        ).tolist()
        coord_names = np.array(
            list(dict(montage.get_positions()["ch_pos"]).keys())
        ).tolist()
    else:
        coord_list = None
        coord_names = None

    return coord_list, coord_names


def read_grid(PATH_GRIDS: _PathLike | None, grid_str: str) -> "pd.DataFrame":
    """Read grid file from path or PYNM_DIR

    Parameters
    ----------
    PATH_GRIDS : _PathLike | None
        path to grid file, by default None
    grid_str : str
        grid name

    Returns
    -------
    pd.DataFrame
        pd.DataFrame including mni x,y,z coordinates for each grid point
    """
    import pandas as pd

    if PATH_GRIDS is None:
        grid = pd.read_csv(PYNM_DIR / ("grid_" + grid_str.lower() + ".tsv"), sep="\t")
    else:
        grid = pd.read_csv(
            PurePath(PATH_GRIDS, "grid_" + grid_str.lower() + ".tsv"), sep="\t"
        )
    return grid


def get_annotations(PATH_ANNOTATIONS: str, PATH_RUN: str, raw_arr: "mne_io.RawArray"):
    filepath = PurePath(PATH_ANNOTATIONS, PurePath(PATH_RUN).name[:-5] + ".txt")
    from mne import read_annotations

    try:
        annot = read_annotations(filepath)
        raw_arr.set_annotations(annot)

        # annotations starting with "BAD" are omitted with reject_by_annotations 'omit' param
        annot_data = raw_arr.get_data(reject_by_annotation="omit")
    except FileNotFoundError:
        logger.critical(f"Annotations file could not be found: {filepath}")

    return annot, annot_data, raw_arr


def read_plot_modules(
    PATH_PLOT: _PathLike = PYNM_DIR / "plots",
):
    """Read required .mat files for plotting

    Parameters
    ----------
    PATH_PLOT : regexp, optional
        path to plotting files, by default
    """

    faces = loadmat(PurePath(PATH_PLOT, "faces.mat"))
    vertices = loadmat(PurePath(PATH_PLOT, "Vertices.mat"))
    grid = loadmat(PurePath(PATH_PLOT, "grid.mat"))["grid"]
    stn_surf = loadmat(PurePath(PATH_PLOT, "STN_surf.mat"))
    x_ver = stn_surf["vertices"][::2, 0]
    y_ver = stn_surf["vertices"][::2, 1]
    x_ecog = vertices["Vertices"][::1, 0]
    y_ecog = vertices["Vertices"][::1, 1]
    z_ecog = vertices["Vertices"][::1, 2]
    x_stn = stn_surf["vertices"][::1, 0]
    y_stn = stn_surf["vertices"][::1, 1]
    z_stn = stn_surf["vertices"][::1, 2]

    return (
        faces,
        vertices,
        grid,
        stn_surf,
        x_ver,
        y_ver,
        x_ecog,
        y_ecog,
        z_ecog,
        x_stn,
        y_stn,
        z_stn,
    )


def write_csv(df, path_out):
    """
    Function to save Pandas dataframes to disk as CSV using
    PyArrow (almost 10x faster than Pandas)
    Difference with pandas.df.to_csv() is that it does not
    write an index column by default
    """
    from pyarrow import csv, Table

    csv.write_csv(Table.from_pandas(df), path_out)


def save_channels(
    nmchannels: "pd.DataFrame",
    out_dir: _PathLike = "",
    prefix: str = "",
) -> None:
    out_dir = Path.cwd() if not out_dir else Path(out_dir)
    filename = "channels.csv" if not prefix else prefix + "_channels.csv"
    write_csv(nmchannels, out_dir / prefix / filename)
    logger.info(f"{filename} saved to {out_dir}")


def save_features(
    df_features: "pd.DataFrame",
    out_dir: _PathLike = "",
    prefix: str = "",
) -> None:
    out_dir = Path.cwd() if not out_dir else Path(out_dir)
    filename = f"{prefix}_FEATURES.csv" if prefix else "_FEATURES.csv"
    write_csv(df_features, out_dir / filename)
    logger.info(f"{filename} saved to {str(out_dir)}")


def save_sidecar(
    sidecar: dict,
    out_dir: _PathLike = "",
    prefix: str = "",
) -> None:
    save_general_dict(sidecar, out_dir, prefix, "_SIDECAR.json")


def save_general_dict(
    dict_: dict,
    out_dir: _PathLike = "",
    prefix: str = "",
    str_add: str = "",
) -> None:
    out_dir = Path.cwd() if not out_dir else Path(out_dir)
    filename = f"{prefix}{str_add}"

    with open(out_dir / prefix / filename, "w") as f:
        json.dump(
            dict_,
            f,
            default=default_json_convert,
            indent=4,
            separators=(",", ": "),
        )
    logger.info(f"{filename} saved to {out_dir}")


def default_json_convert(obj) -> list | float:
    import pandas as pd

    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, pd.DataFrame):
        return obj.to_numpy().tolist()
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    raise TypeError("Not serializable")


def read_sidecar(PATH: _PathLike) -> dict:
    with open(PurePath(str(PATH) + "_SIDECAR.json")) as f:
        return json.load(f)


def read_features(PATH: _PathLike) -> "pd.DataFrame":
    import pandas as pd

    return pd.read_csv(str(PATH) + "_FEATURES.csv", engine="pyarrow")


def read_channels(PATH: _PathLike) -> "pd.DataFrame":
    import pandas as pd

    return pd.read_csv(str(PATH) + "_channels.csv")


def get_run_list_indir(PATH: _PathLike) -> list:
    from os import walk

    f_files = []
    # for dirpath, _, files in Path(PATH).walk(): # Only works in python >=3.12
    for dirpath, _, files in walk(PATH):
        for x in files:
            if "FEATURES" in x:
                f_files.append(PurePath(dirpath).name)
    return f_files


def loadmat(filename) -> dict:
    """
    this function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    """
    from scipy.io import loadmat as sio_loadmat

    data = sio_loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


def get_paths_example_data():
    """
    This function should provide RUN_NAME, PATH_RUN, PATH_BIDS, PATH_OUT and datatype for the example
    dataset used in most examples.
    """

    sub = "testsub"
    ses = "EphysMedOff"
    task = "gripforce"
    run = 0
    datatype = "ieeg"

    # Define run name and access paths in the BIDS format.
    RUN_NAME = f"sub-{sub}_ses-{ses}_task-{task}_run-{run}"

    PATH_BIDS = PYNM_DIR / "data"

    PATH_RUN = PYNM_DIR / "data" / f"sub-{sub}" / f"ses-{ses}" / datatype / RUN_NAME

    # Provide a path for the output data.
    PATH_OUT = PATH_BIDS / "derivatives"

    return RUN_NAME, PATH_RUN, PATH_BIDS, PATH_OUT, datatype


def _check_keys(dict):
    """
    checks if entries in dictionary are mat-objects. If yes
    todict is called to change them to nested dictionaries
    """
    from scipy.io.matlab import mat_struct

    for key in dict:
        if isinstance(dict[key], mat_struct):
            dict[key] = _todict(dict[key])
    return dict


def _todict(matobj) -> dict:
    """
    A recursive function which constructs from matobjects nested dictionaries
    """
    from scipy.io.matlab import mat_struct

    dict = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, mat_struct):
            dict[strg] = _todict(elem)
        else:
            dict[strg] = elem
    return dict


def generate_unique_filename(path: _PathLike):
    path = Path(path)

    dir = path.parent
    filename = path.stem
    extension = path.suffix

    counter = 1
    while True:
        new_filename = f"{filename}_{counter}{extension}"
        new_file_path = dir / new_filename
        if not new_file_path.exists():
            return Path(new_file_path)
        counter += 1



import os
import platform
from pathlib import PurePath
from importlib.metadata import version
from py_neuromodulation.utils.logging import NMLogger


#####################################
# Globals and environment variables #
#####################################

__version__ = version("py_neuromodulation")

# Check if the module is running headless (no display) for tests and doc builds
PYNM_HEADLESS: bool = not os.environ.get("DISPLAY")
PYNM_DIR = PurePath(__file__).parent  # Define constant for py_nm directory

os.environ["MPLBACKEND"] = "agg" if PYNM_HEADLESS else "qtagg"  # Set matplotlib backend

os.environ["LSLAPICFG"] = str(PYNM_DIR / "lsl_api.cfg")  # LSL config file


# Set  environment variable MNE_LSL_LIB (required to import Stream below)
LSL_DICT = {
    "windows_32bit": "windows/x86/liblsl.1.16.2.dll",
    "windows_64bit": "windows/amd64/liblsl.1.16.2.dll",
    "darwin_i386": "macos/amd64/liblsl.1.16.2.dylib",
    "darwin_arm": "macos/arm64/liblsl.1.16.0.dylib",
    "linux_jammy_32bit": "linux/jammy_x86/liblsl.1.16.2.so",
    "linux_jammy_64bit": "linux/jammy_amd64/liblsl.1.16.2.so",
    "linux_focal_64bit": "linux/focal_amd64/liblsl.1.16.2.so",
    "linux_bionic_64bit": "linux/bionic_amd64/liblsl.1.16.2.so",
    "linux_bookworm_64bit": "linux/bookworm_amd64/liblsl.1.16.2.so",
    "linux_noble_64bit": "linux/noble_amd64/liblsl.1.16.2.so",
    "linux_32bit": "linux/jammy_x86/liblsl.1.16.2.so",
    "linux_64bit": "linux/jammy_amd64/liblsl.1.16.2.so",
}

PLATFORM = platform.system().lower().strip()
ARCH = platform.architecture()[0]

match PLATFORM:
    case "windows":
        KEY = PLATFORM + "_" + ARCH
    case "darwin":
        KEY = PLATFORM + "_" + platform.processor()
    case "linux":
        if "VERSION_CODENAME" in platform.freedesktop_os_release().keys():
            DIST = platform.freedesktop_os_release()["VERSION_CODENAME"]
            KEY = PLATFORM + "_" + DIST + "_" + ARCH
        else:
            KEY = PLATFORM + "_" + ARCH
    case _:
        KEY = ""

if KEY in LSL_DICT:
    os.environ["MNE_LSL_LIB"] = str(PYNM_DIR / "liblsl" / LSL_DICT[KEY])


# Create global dictionary to store user-defined features
user_features = {}


######################################
# Logger initialization and settings #
######################################

logger = NMLogger(__name__)  # logger initialization first to prevent circular import

####################################
# API: Exposed classes and methods #
####################################
from .stream.stream import Stream
from .stream.data_processor import DataProcessor
from .stream.settings import NMSettings

from .analysis.feature_reader import FeatureReader
from .features.feature_processor import add_custom_feature, remove_custom_feature

from .utils import types
from .utils import io

from . import stream
from . import analysis

from .stream.settings import get_default_settings, get_fast_compute, reset_settings

from .gui.backend.app_manager import AppManager as App



import webbrowser

from py_neuromodulation.stream import (
    LSLStream,
    LSLOfflinePlayer,
)
from py_neuromodulation import io
from py_neuromodulation import App


def main():
    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = io.get_paths_example_data()

    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = io.read_BIDS_data(PATH_RUN=PATH_RUN)

    player = LSLOfflinePlayer(raw=raw, stream_name="example_stream")

    player.start_player(chunk_size=30, n_repeat=5999999)

    App(run_in_webview=False, dev=False, debug=False).launch()

if __name__ == "__main__":
    main()



import numpy as np
import pytest

import py_neuromodulation as nm


def test_stream_init():
    """Test if stream initialization with passed data will setup channels correctly"""
    np.random.seed(0)
    data = np.random.random((10, 1000))
    sfreq = 1000
    stream = nm.Stream(sfreq=sfreq, data=data, sampling_rate_features_hz=11)

    assert stream.channels.shape[0] == 10
    assert stream.settings.sampling_rate_features_hz == 11


def test_stream_init_no_sfreq():
    """Check if stream initialization without sfreq will raise an error"""
    np.random.seed(0)
    data = np.random.random((10, 1000))
    with pytest.raises(Exception):
        nm.Stream(data=data, sampling_rate_features_hz=11)


def test_init_warning_no_used_channel():
    """Check if a warning is raised when a stream is initialized with channels, but no row has used == 1 and target == 0"""
    np.random.seed(0)
    data = np.random.random((10, 1000))
    sfreq = 1000
    stream = nm.Stream(sfreq=sfreq, data=data, sampling_rate_features_hz=11)
    channels = stream.channels
    channels["used"] = 0

    with pytest.raises(Exception):
        nm.Stream(
            sfreq=sfreq,
            data=data,
            channels=channels,
            sampling_rate_features_hz=11,
        )
